# trees_addremoverings.tcl - Adding and removing ring boundaries.

# Procedure Index:
# proc add_ring { var1 var2 dlg imagewin zoomwin plotwin top timeseries series bottom scroll }
# proc remove_ring { var1 var2 dlg imagewin zoomwin plotwin top timeseries series bottom scroll }

# If necessary reload the lists of chain-coded boundaries, the image mosaic,
# and its index, from the files stored in the directory given by sdir.

proc reload_persistent_data { sdir } {
    global findrings mosaic

    if {![info exists findrings(loaded,chainmap)]} {
        set findrings(loaded,chainmap) \
            [Sadie_TREES_Disk2Chain \
                 $sdir/images/mosaic/$findrings(file,chainmap)]
    }
    if {![info exists findrings(loaded,smallchainmap)]} {
        set findrings(loaded,smallchainmap) \
            [Sadie_TREES_Disk2Chain \
                 $sdir/images/mosaic/$findrings(file,smallchainmap)]
    }
    if {![info exists mosaic(loaded,mosaic)]} {
        set mosaic(loaded,mosaic) \
            [Sadie_TREES_ByteOpen \
                 $sdir/images/mosaic/$mosaic(file,mosaic)]
    }
    if {![info exists mosaic(loaded,index)]} {
        set mosaic(loaded,index) \
            [Sadie_TREES_MosaicIndexOpen \
                 $sdir/images/mosaic/$mosaic(file,index)]
    }

}

# Display the dialog for choosing betweem the browse window and the zoom window
# for boundary addition.

proc pick_window_for_add { dlg imagewin zoomwin } {
    global zoom

    if { $zoom(frozen) } {         
        set answer [tk_dialog $dlg "Select Add Ring Location" \
                        "Click Add Ring, then click in the window where ring is to be added, or\nClick Cancel to quit this operation." \
                        "" 0 "Add Ring in\nBrowse Window" \
                        "Add Ring in\nZoom Window" \
                        "Cancel\nAdd Ring"]
        switch -exact -- $answer {
            0 { return $imagewin }
            1 { return $zoomwin }
            2 { return "" }
        }   
    } else {
        set answer [tk_dialog $dlg "Select Add Ring Location" \
                        "Click Add Ring, then click on image where ring is to be added.\nClick Cancel to quit this operation." \
                        "" 0 "Add Ring in Browse Window" \
                        "Cancel Add Ring"]
        switch -exact -- $answer {
            0 { return $imagewin }
            1 { return "" }
        }            
    }
}

# Display the dialog for choosing betweem the browse window and the zoom window
# for boundary removal.

proc pick_window_for_remove { dlg imagewin zoomwin } {
    global zoom

    if { $zoom(frozen) } {         
        set answer [tk_dialog $dlg "Select Ring Boundary for Removal" \
                        "Click Remove Boundary, then click in the window containing the ring boundary to remove, or\nClick Cancel to quit this operation." \
                        "" 0 "Remove Boundary in\nBrowse Window" \
                        "Remove Boundary in\nZoom Window" \
                        "Cancel"]
        switch -exact -- $answer {
            0 { return $imagewin }
            1 { return $zoomwin }
            2 { return "" }
        }   
    } else {
        set answer [tk_dialog $dlg "Select Ring Boundary for Removal" \
                        "Click Remove Boundary, then click on the boundary to be removed.\nClick Cancel to quit this operation." \
                        "" 0 "Remove Boundary" \
                        "Cancel"]
        switch -exact -- $answer {
            0 { return $imagewin }
            1 { return "" }
        }            
    }
}

# Display a provisionally modified ring boundary, using both the browse and
# zoom canvases if necessary.

proc display_provisional { imagewin zoomwin hashdoc provbnum } {
    global zoom boundarycolor

    TREES_display_small_provisional $imagewin $hashdoc \
        $provbnum $boundarycolor(added)
    if { $zoom(frozen) } {         
        TREES_display_provisional $zoomwin $zoom(X1) $zoom(Y1) \
            $hashdoc $provbnum $boundarycolor(added)
    }
}

# Highlight the ring boundary the user has selected for removal.

proc color_removed { imagewin zoomwin hashdoc provbnum } {
    global zoom boundarycolor

    TREES_color_provisional $imagewin $hashdoc $provbnum \
        $boundarycolor(removed)
    if { $zoom(frozen) } {         
        TREES_color_provisional $zoomwin $hashdoc $provbnum \
            $boundarycolor(removed)
    }
}

# Re-draw a normal ring boundary.

proc display_normal { imagewin zoomwin hashdoc provbnum } {
    global zoom boundarycolor

    TREES_display_small_provisional $imagewin $hashdoc \
        $provbnum $boundarycolor(normal)
    if { $zoom(frozen) } {         
        TREES_display_provisional $zoomwin $zoom(X1) $zoom(Y1) \
            $hashdoc $provbnum $boundarycolor(normal)
    }
}

# Remove any color highlighting a previously selected boundary.

proc color_normal { imagewin zoomwin hashdoc provbnum } {
    global zoom boundarycolor

    TREES_color_provisional $imagewin $hashdoc $provbnum \
        $boundarycolor(normal)
    if { $zoom(frozen) } {         
        TREES_color_provisional $zoomwin $hashdoc $provbnum \
            $boundarycolor(normal)
    }
}

# Delete a previously drawn provisional ring from the canvases.

proc undisplay_provisional { imagewin zoomwin hashdoc provbnum } {
    global zoom

    TREES_undisplay_provisional $imagewin $hashdoc $provbnum
    if { $zoom(frozen) } {
        TREES_undisplay_provisional $zoomwin $hashdoc $provbnum
    }    
}

# Completely re-draw the brightness profile plots.

proc refresh_profiles { } {
    global measurements gui

    if {$gui(profileplot) != 0} {
        trees_destroy_profileplot .trees_main.profilepanel
    }
    draw_profileplot .trees_main.profilepanel $measurements(profiles)
    set gui(profileplot) 1
}

# Update all the derived data (generally after committing a change to the
# primary measurement and housekeeping data structures).

proc refresh_data { sdir chainmap smallchainmap series skelname plotwin top \
                        timeseries series bottom scroll imagewin } {

    global measurements findrings
    upvar #0 $skelname SkelPlot
    upvar #0 $SkelPlot(sample,$series,masterlist) masterlist
    
    Sadie_TREES_Chain2Disk $chainmap \
        $sdir/images/mosaic/$findrings(file,chainmap)
    Sadie_TREES_Chain2Disk $smallchainmap \
        $sdir/images/mosaic/$findrings(file,smallchainmap) 
    compute_indices $skelname $SkelPlot(sample,$series,masterlist) \
        $SkelPlot(sample,$series,serieslist) $series
    compute_bones $skelname $SkelPlot(sample,$series,serieslist)
    redraw_skelplot $skelname $plotwin $top $timeseries \
        $series $bottom $scroll $imagewin
    refresh_profiles
    incr SkelPlot(sample,$series,addBoundaryNum)
    trees_session_save
}

# Commit a provisionally inserted ring boundary to the primary measurement
# and housekeeping data structures.

proc commit_provisional { sdir chainmap smallchainmap skelname \
                              imagewin zoomwin plotwin top timeseries series \
                              bottom scroll provbnum } {

    global measurements findrings skelplot
    upvar #0 $skelname SkelPlot
    upvar #0 $SkelPlot(sample,$series,masterlist) masterlist
    
    display_normal $imagewin $zoomwin $sdir $provbnum
    set newindex [TREES_commit_provisional \
                      $sdir $provbnum $chainmap $smallchainmap \
                      measurements(ringwidth) measurements(profiles) \
                      measurements(diff_profiles) measurements(earlywidth) \
                      measurements(latewidth) measurements(boundbox) \
                      masterlist skelplot(notes)]
    refresh_data $sdir $chainmap $smallchainmap $series $skelname $plotwin \
        $top $timeseries $series $bottom $scroll $imagewin
}

# Commit a provisionally removed ring boundary to the primary measurement
# and housekeeping data structures.

proc commit_removal { sdir chainmap smallchainmap skelname imagewin zoomwin \
                          plotwin top timeseries series bottom scroll \
                          provbnum } {

    global measurements findrings skelplot
    upvar #0 $skelname SkelPlot
    upvar #0 $SkelPlot(sample,$series,masterlist) masterlist
    
    undisplay_provisional $imagewin $zoomwin $sdir $provbnum
    set newindex [TREES_commit_removal \
                      $sdir $provbnum $chainmap $smallchainmap \
                      measurements(ringwidth) measurements(profiles) \
                      measurements(diff_profiles) measurements(earlywidth) \
                      measurements(latewidth) measurements(boundbox) \
                      masterlist skelplot(notes)]
    refresh_data $sdir $chainmap $smallchainmap $series $skelname $plotwin \
        $top $timeseries $series $bottom $scroll $imagewin
}

# Add a new ring boundary to the sequence of previously determined boundaries.

proc add_ring { var1 var2 dlg imagewin zoomwin plotwin top timeseries \
                    series bottom scroll } {

    global zoom findrings boundarycolor session measurements mosaic gui
    
    upvar #0 $var1 SkelPlot
    upvar #0 $var2 Image
    
    set sdir $session(directory)
    reload_persistent_data $sdir
    set zoomfactor [expr $zoom(default) / $findrings(ring,resolution)]
    set adding_rings true
    while {$adding_rings} {
        set provbnum $SkelPlot(sample,$series,addBoundaryNum)
        set grabWin [pick_window_for_add $dlg $imagewin $zoomwin]
        set adding_rings [expr {$grabWin ne ""}]
        set selecting_boundary $adding_rings
        while {$adding_rings && $selecting_boundary} {   
            grab set $grabWin
            vwait Image(location)
            grab release $grabWin
            set x [lindex $Image(location) 0]
            set y [lindex $Image(location) 1]         
            if { $grabWin == $zoomwin } {
                incr x $zoom(X1)
                incr y $zoom(Y1)
                if { $x < $zoom(X1) || $x > $zoom(X2) \
                         || $y < $zoom(Y1) || $y > $zoom(Y2) } {
                    set answer [tk_dialog $dlg "Unable to Add Ring" \
                                    "You must click inside the Zoom Window" \
                                    "" 0 "Try Again" "Cancel"]
                    set selecting_boundary [expr $answer == 0]
                    set adding_rings [expr $answer != 1]
                    continue
                }
            } else {
                set x [expr round(($x + ([lindex [$imagewin xview] 0] \
                                             * [$imagewin cget -width])) \
                                      * $zoomfactor)]
                set y [expr round(($y + ([lindex [$imagewin yview] 0] \
                                             * [$imagewin cget -height])) \
                                      * $zoomfactor)]
            }
            if {[TREES_add_provisional $sdir $provbnum $mosaic(loaded,mosaic) \
                     $mosaic(loaded,index) $findrings(loaded,chainmap) \
                     $zoomfactor $x $y measurements(boundbox)]} {
                 display_provisional $imagewin $zoomwin $sdir $provbnum
                 set answer [tk_dialog $dlg "Confirm Add Ring Location" \
                                 "Is this location OK\? " \
                                 "" 0 "OK" "Try Again" "Cancel Add Ring"]
                 undisplay_provisional $imagewin $zoomwin $sdir $provbnum
                 set selecting_boundary  [expr $answer == 1]
                 set adding_rings  [expr $answer != 2]
             } else {
                 set answer [tk_dialog $dlg "Unable to Add Ring" \
                                 "Unable to insert a new ring boundary at this position." \
                                 "" 0 "Try Again" "Cancel"]
                 set selecting_boundary [expr $answer == 0]
                 set adding_rings [expr $answer != 1]
             }
        }
        if { $adding_rings } {
            commit_provisional $sdir $findrings(loaded,chainmap) \
                $findrings(loaded,smallchainmap) $var1 $imagewin $zoomwin \
                $plotwin $top $timeseries $series $bottom $scroll $provbnum
        }
    }
}

# Remove an existing ring boundary from the sequence of previously determined
# boundaries.

proc remove_ring { var1 var2 dlg imagewin zoomwin plotwin top timeseries \
                       series bottom scroll } {

    global zoom findrings boundarycolor session measurements mosaic gui
    
    upvar #0 $var1 SkelPlot
    upvar #0 $var2 Image
    
    set sdir $session(directory)
    reload_persistent_data $sdir
    set zoomfactor [expr $zoom(default) / $findrings(ring,resolution)]
    set removing_rings true
    while {$removing_rings} {
        set provbnum $SkelPlot(sample,$series,addBoundaryNum)
        set grabWin [pick_window_for_remove $dlg $imagewin $zoomwin]
        set removing_rings [expr {$grabWin ne ""}]
        set selecting_boundary $removing_rings
        while {$removing_rings && $selecting_boundary} {   
            grab set $grabWin
            vwait Image(location)
            grab release $grabWin
            set x [lindex $Image(location) 0]
            set y [lindex $Image(location) 1]         
            if { $grabWin == $zoomwin } {
                incr x $zoom(X1)
                incr y $zoom(Y1)
                if { $x < $zoom(X1) || $x > $zoom(X2) \
                         || $y < $zoom(Y1) || $y > $zoom(Y2) } {
                    set answer [tk_dialog $dlg "Unable to Remove Ring Boundary" \
                                    "You must click inside the Zoom Window" \
                                    "" 0 "Try Again" "Cancel"]
                    set selecting_boundary [expr $answer == 0]
                    set removing_rings [expr $answer != 1]
                    continue
                }
            } else {
                set x [expr round(($x + ([lindex [$imagewin xview] 0] \
                                             * [$imagewin cget -width])) \
                                      * $zoomfactor)]
                set y [expr round(($y + ([lindex [$imagewin yview] 0] \
                                             * [$imagewin cget -height])) \
                                      * $zoomfactor)]
            }
            if {[TREES_remove_provisional $sdir $provbnum $mosaic(loaded,mosaic) \
                     $mosaic(loaded,index) $findrings(loaded,chainmap) \
                     $findrings(loaded,smallchainmap) $x $y \
                     measurements(boundbox)]} {
                 color_removed $imagewin $zoomwin $sdir $provbnum
                 set answer [tk_dialog $dlg "Confirm Remove Ring Location" \
                                 "Is this location OK\? " \
                                 "" 0 "OK" "Try Again" "Cancel Remove Ring"]
                 color_normal $imagewin $zoomwin $sdir $provbnum
                 set selecting_boundary  [expr $answer == 1]
                 set removing_rings  [expr $answer != 2]
             } else {
                 set answer [tk_dialog $dlg "Unable to Remove Ring" \
                                 "Unable to remove an existing ring boundary at this position." \
                                 "" 0 "Try Again" "Cancel"]
                 set selecting_boundary [expr $answer == 0]
                 set removing_rings [expr $answer != 1]
             }
        }
        if { $removing_rings } {
            commit_removal $sdir $findrings(loaded,chainmap) \
                $findrings(loaded,smallchainmap) $var1 $imagewin $zoomwin \
                $plotwin $top $timeseries $series $bottom $scroll $provbnum
        }
    }
}
