#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <tcl.h>
#include <tk.h>
#include <sadie.h>
#include "proto.h"
#include "sadie_byte.h"
#include "trees.h"
#include "mosaic.h"
#include "chain.h"
#include "histogram.h"
#include <sys/time.h>
#if WITH_DMALLOC
#include <dmalloc.h>
#endif /* WITH_DMALLOC */


/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   Global Sadie variables that must be set from within Tcl/Tk.              */
/*                                                                            */
/*                                                                            */
/*----------------------------------------------------------------------------*/
extern 	short   nlev;
extern	short   csize;
extern	double  weight;
extern	double  *count;
extern	PIXEL   gain;
extern	PIXEL   bias;
extern	PIXEL   gmin;
extern	PIXEL   gmax;
extern	PIXEL   thresh;
extern	PIXEL   gbrk[2][4];
extern	PIXEL   *table;


/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   Variables required for analyst interaction.                              */
/*                                                                            */
/*                                                                            */
/*----------------------------------------------------------------------------*/
int  locx;      /* Column coordinate of the point of user click.              */
int  locy;      /* Row coordinate of the point of user click.                 */
int  zoomx;     /* Column coordinate of top-left corner of zoom window.       */
int  zoomy;     /* Row coordinate of top-left corner of zoom window.          */
int  retain;    /* Variable that tells whether or not to retain the ring.     */


/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   Variables necessary for computing vstart and vend with bresh algorithm.  */
/*                                                                            */
/*                                                                            */
/*----------------------------------------------------------------------------*/
extern int **x_array_buffer; /* The data where info is stored.                */
extern int point;            /* Counter for points used.                      */
extern int count_point;      /* Flag to decide whether to count or fill array.*/


/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This procedure finds the rings in the mosaic image.                      */
/*                                                                            */
/*                                                                            */
/*----------------------------------------------------------------------------*/
int Sadie_TREES_Ext_TrackRingsCmd(ClientData client_data, Tcl_Interp* interp, int argc, char *argv[])
{
    /* General declarations */
    int   addrstr;
    char  msg[SLEN];
    char *mode = NULL;
    struct  timeval start, end, t1, t2;
    HISTOGRAM  globalhist;
    HISTOGRAM  **localhist=NULL;
    IMAGE_BYTE  *byteimg=NULL;
    IMAGE_BYTE  *iMag=NULL;
    IMAGE_BYTE  *iDir=NULL;
    MOSAIC_INDEX  *index=NULL;
    
    /* Declarations for edge tracking go in here */
    register int i, j, k, l, m, n;
    int  max, found_max, done, nmax;
    int  startx, starty;
    IMAGE_BYTE  *out = NULL;

    /* Declarations for chain-coding go in here */
    int  discard, x, y, count;
    int  numlinks, pixcount, maxx, maxy;
    int  prevx, prevy, currx, curry, reset;
    unsigned char  code;
    CHAIN  *chainmap;
    LIST_NODE  *tempLnode;
    LIST_NODE  **prevLnode;
    LIST_NODE  *traverseLnode;
    CHAIN_NODE  **prevCnode;
    CHAIN_NODE  *tempCnode;


    /* Get the arguments */
    if( argc <= 3 ) {
	sprintf(msg, "Usage: Sadie_TREES_TrackRings byteimg index mode");
	Tcl_SetResult(interp, msg, TCL_VOLATILE);
	return TCL_ERROR;
    }
    
    sscanf(argv[1], "%x", &addrstr);
    byteimg = (IMAGE_BYTE *) addrstr;
    
    sscanf(argv[2], "%x", &addrstr);
    index = (MOSAIC_INDEX *) addrstr;

    mode = argv[3];

    /* start timing */
    gettimeofday(&start, NULL);
    
    /* This will return the gradient magnitude and direction mosaics */
    gettimeofday(&t1, NULL);
    MOSAIC_BYTE_TRACKRINGS(byteimg, index, &iMag, &iDir);
    gettimeofday(&t2, NULL);
    printf("Total time for gradient = %ld ms.\n\n", delay(t1, t2));
    
    /* Now we'll suppress noise edge pixels */
    printf("Post-gradient processing\n");
    gettimeofday(&t1, NULL);
    MOSAIC_BYTE_SUPPRESS(byteimg, index, iMag, iDir, &globalhist, &localhist);
    gettimeofday(&t2, NULL);
    printf("Total post-gradient time = %ld ms.\n\n", delay(t1, t2));


    /* check input image */
    if (!CHECKIMG_BYTE(byteimg)) {
	sprintf(msg," Can't identify mosaic.");
	Tcl_SetResult(interp, msg, TCL_VOLATILE);
	return TCL_ERROR;
    }
    if (!CHECKIMG_BYTE(iMag)) {
	sprintf(msg," Can't identify magnitude mosaic.");
	Tcl_SetResult(interp, msg, TCL_VOLATILE);
	return TCL_ERROR;
    }
    if (!CHECKIMG_BYTE(iDir)) {
	sprintf(msg," Can't identify direction mosaic.");
	Tcl_SetResult(interp, msg, TCL_VOLATILE);
	return TCL_ERROR;
    }
    if (!CHECKIMG_MOSAIC_INDEX(index)) {
	sprintf(msg," Can't identify mosaic index");
	Tcl_SetResult(interp, msg, TCL_VOLATILE);
	return TCL_ERROR;
    }
    if ((iMag->nlin != iDir->nlin) || (iMag->npix != iDir->npix)) {
	sprintf(msg," Dimensions of Ring Mosaic and Link Mosaic don't match.");
	Tcl_SetResult(interp, msg, TCL_VOLATILE);
	return TCL_ERROR;
    }


    gettimeofday(&t1, NULL);
    
    /* Allocate memory for storing the ring image */
    GETMEM_BYTE(iMag->nbnd, iMag->nlin, iMag->npix, &out);
    
    /* Allocate memory for chainmap */
    chainmap = (CHAIN *)malloc(sizeof(CHAIN));
    
    /* Initialize all the pixels to zero */
    for(i = 0; i < out->nbnd; i++)
	for(j = 0; j < out->nlin; j++)
	    for(k = 0; k < out->npix; k++)
		out->data[i][j][k] = (BYTE)0;
    
    
    /* Write chain header */
    chainmap->nlin = index->nlin;
    chainmap->npix = index->npix;
    memcpy((char *)(chainmap->header), (char *)byteimg->text, TLEN - 1);
    chainmap->list = NULL;
    prevLnode = (LIST_NODE **)&(chainmap->list);


    /* Edge tracking */
    max = 5;
    found_max = 0;
    count = -1;
    for(k = index->hstart + 1; k <= index->hend; k++) {
	j = index->vstart[k] - index->voffset[k] + 5;

	/* Identify starting point of a ring */
	if (iMag->data[0][j][k] > max) {
	    max = iMag->data[0][j][k];
	    startx = index->vstart[k] + 0;
	    starty = k;
	    found_max = 1;
	}
	
	/* Start tracking the ring */
	if (found_max && iMag->data[0][j][k] <= 5) {
	    found_max = 0;
	    max = 5;
	    done = 0;
	    discard = numlinks = 0;
	    pixcount = 1;

	    tempLnode = (LIST_NODE *)malloc(sizeof(LIST_NODE));
	    tempLnode->startj = startx;
	    tempLnode->startk = starty;
	    tempLnode->ring_type = NORMAL;
	    tempLnode->chain = NULL;
	    tempLnode->next = NULL;
	    prevCnode = (CHAIN_NODE **)&(tempLnode->chain);

	    /* For as long as the ring is not fully coded */
	    while (!done && !discard) {
		    
		/* Check for the end of ROI */
		if (startx == index->vend[starty] - 5 ||
		    starty == index->hstart + 5 || starty == index->hend - 5) {
		    done = 1;
		    break;
		}
		
		/* Find the neighbour with the highest gradient magnitude */
		if (HIST_FIND_MAX_NEIGH(iMag, byteimg, index, startx, starty, &x, &y)) {
		    
		    /* Check for an intersection with an already coded ring */
		    if (out->data[0][x-index->voffset[y]][y] == 255) {
			discard = 1;
			break;
			
		    } else { /* No intersection, then we code the pixel */

			out->data[0][x - index->voffset[y]][y] = 255;
			CHAIN_COMPUTE_CODE(startx, starty, x, y, &code);
			
			startx = x;
			starty = y;
			tempCnode = (CHAIN_NODE *)malloc(sizeof(CHAIN_NODE));
			tempCnode->code = code;
			tempCnode->valid = 1;
			tempCnode->next = NULL;
			(*prevCnode) = tempCnode;
			prevCnode = (CHAIN_NODE **)&(tempCnode->next);
			pixcount++;
			
		    }
		    
		    
		} else { /* No neighbours */

		    if ( (float)numlinks/pixcount > 0.25 && pixcount > 15) { 
			/* Too many links....discard ring */
			discard = 1;
			break;
			
		    } else { /* Try bridging a long gap */

			nmax = 0;
			for(m = starty - 5; m <= starty + 5; m++) {
			    if (m < index->hstart || m > index->hend)
				continue;
			    for(n = startx + 1; n <= startx + 10; n++) {
				if (n < index->vstart[m] || n > index->vend[m])
				    continue;
				if (iMag->data[0][n - index->voffset[m]][m] > nmax) {
				    nmax = iMag->data[0][n - index->voffset[m]][m];
				    maxx = n;
				    maxy = m;
				}
			    }
			}
			
			if (nmax == 0) { /* If unable to brigde */
			    
			    /* End if the current pixel is close to the frame boundary */
			    if (startx >= index->vend[starty] - 5 ||
				starty == index->hstart + 5 || starty == index->hend - 5) {
				done = 1;
				break;
			    } else if (mode[0] == 'I' && pixcount > 50) { /* User Interaction */

				/* Call proc to display region in zoom panel */
				sprintf(msg, "interaction_zoom %d %d", starty, startx);
				if (Tcl_Eval(interp, msg) != TCL_OK)
				    return TCL_ERROR;

				/* Decode the chain so far and display on zoom window */
				traverseLnode = tempLnode;
				prevx = traverseLnode->startk;
				prevy = traverseLnode->startj;
				tempCnode = (CHAIN_NODE *)traverseLnode->chain;
				while(tempCnode != NULL) {

				    CHAIN_COMPUTE_COORDINATES(prevx, prevy, tempCnode->code, &currx, &curry);
				    sprintf(msg, "%s create line %d %d %d %d -width 1 -fill %s -tag %s", ".trees_main.zoompanel.zoomframe.canvas", prevx-zoomx, prevy-zoomy, currx-zoomx, curry-zoomy, "red", "interact");

				    if (Tcl_Eval(interp, msg) != TCL_OK)
					return TCL_ERROR;

				    prevx = currx;
				    prevy = curry;
				    tempCnode = (CHAIN_NODE *)tempCnode->next;
				}

				/* Make sure the ring is visible */
				sprintf(msg, "update idletasks");
				if (Tcl_Eval(interp, msg) != TCL_OK)
				    return TCL_ERROR;

				/* Prompt and wait for user input */
				sprintf(msg, "trees_interaction_init");
				if (Tcl_Eval(interp, msg) != TCL_OK)
				    return TCL_ERROR;

				/* Clear up the zoom window */
				sprintf(msg, "catch {.trees_main.zoompanel.zoomframe.canvas delete all}");
				
				maxy = locx + zoomx;
				maxx = locy + zoomy;
				
				/* Discard on analyst instruction */
				if (!retain) {
				    discard = 1;
				    break;
				}
				
			    } else {
				discard = 1;
				break;
			    }
			}
			
			/* Count number of points in line */
			count_point = TRUE;
			point = 0;
			brshnm(starty, startx, maxy, maxx);
			
			/* Allocate memory */
			x_array_buffer = (int **) malloc((long)point*sizeof(int *));
			for  (l = 0; l < point; l++)
			    x_array_buffer[l] = (int  *) calloc(2, sizeof(int));
			
			/* Fill array */ 
			count_point = FALSE;
			point = 0;
			brshnm(starty, startx, maxy, maxx);
			
			/* Link */
			for (l = 0; l < point; l++) {
			    
			    x = x_array_buffer[l][1];
			    y = x_array_buffer[l][0];
			    out->data[0][x - index->voffset[y]][y] = 255;
			    CHAIN_COMPUTE_CODE(startx, starty, x, y, &code);
			    
			    startx = x;
			    starty = y;
			    tempCnode = (CHAIN_NODE *)malloc(sizeof(CHAIN_NODE));
			    tempCnode->code = code;
			    tempCnode->valid = 0;
			    tempCnode->next = NULL;
			    (*prevCnode) = tempCnode;
			    prevCnode = (CHAIN_NODE **)&(tempCnode->next);
			    pixcount++;
			    numlinks++;
			    
			}
			
			if(x_array_buffer) {      
			  /* Free Bresh Array */
			  for (l = 0; l < point; l++){
			    if(x_array_buffer[l]) {
			      free(x_array_buffer[l]);
			      x_array_buffer[l] = NULL;
			    }
			  }
			  
			  free(x_array_buffer);
			  x_array_buffer = NULL;
			}
			
		    }
		    
		}
		
		
	    } /* While loop */
	    
	    if (!discard && pixcount > 15) { /* Retain ring */
		(*prevLnode) = tempLnode;
		prevLnode = (LIST_NODE **)&(tempLnode->next);
		tempLnode->ring_id = count++;
	    } else {
		/* Reset the current chain to zero */
		do {
		    reset = 0;
		    out->data[0][startx - index->voffset[starty]][starty] = 0;
		    for(l = starty + 1; l > starty - 2; l--) {
			if (l < index->hstart || l > index->hend)
			    continue;
			for(m = startx - 1; m < startx + 1; m++) {
			    if (m < index->vstart[l] || m > index->vend[l])
				continue;
			    if (out->data[0][m - index->voffset[l]][l] == 255) {
				out->data[0][m - index->voffset[l]][l] = 0;
				startx = m;
				starty = l;
				reset = 1;
				break;
			    }
			}
			if (reset) break;
		    }
		    if (!reset) break;
		} while (startx != index->vstart[starty] + 1);
		
	    }
	    
	} /* if condition */
	
    } /* Outer for loop */
    
    gettimeofday(&t2, NULL);
    printf("Tracking and coding rings = %ld ms.\n\n", delay(t1, t2));

    gettimeofday(&end, NULL);
    printf("Total time for finding rings = %ld ms.\n\n", delay(start, end));

    printf("Encoded %d rings\n", count+1);
    
    /*Release the memory*/
    
    if (CHECKIMG_BYTE(out))  RELIMG_BYTE(&out);
    if (CHECKIMG_BYTE(iMag))  RELIMG_BYTE(&iMag);
    if (CHECKIMG_BYTE(iDir))  RELIMG_BYTE(&iDir);
    if (localhist) free(localhist);

    /* Return only the chainmap */
    sprintf(msg, "%x", (int)chainmap);
    Tcl_SetResult(interp, msg, TCL_VOLATILE);

    
    return TCL_OK;
}


/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This procedure conveys the user inputs to the algorithm.                 */
/*                                                                            */
/*                                                                            */
/*----------------------------------------------------------------------------*/
int Sadie_TREES_Ext_SetUserInputCmd(ClientData client_data, Tcl_Interp* interp, int argc, char *argv[])
{
    double  row, col;
    char msg[SLEN];
    
    if( argc <= 5 ) {
	sprintf(msg, "Usage: Sadie_TREES_SetUserInput retain locx locy zoomx zoomy");
	Tcl_SetResult(interp, msg, TCL_VOLATILE);
	return TCL_ERROR;
    }
    
    if( Tcl_GetInt (interp, argv[1], &retain) != TCL_OK ) {
	return TCL_ERROR;
    }
    if( Tcl_GetDouble (interp, argv[2], &col) != TCL_OK ) {
	return TCL_ERROR;
    }
    if( Tcl_GetDouble (interp, argv[3], &row) != TCL_OK ) {
	return TCL_ERROR;
    }
    if( Tcl_GetInt (interp, argv[4], &zoomx) != TCL_OK ) {
	return TCL_ERROR;
    }
    if( Tcl_GetInt (interp, argv[5], &zoomy) != TCL_OK ) {
	return TCL_ERROR;
    }
      
    /* User Input convey protocol */
    /* locx, locy, discard and read are globals */
    locx = (int)col;
    locy = (int)row;
    
    return TCL_OK;
}


/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This procedure initializes the the TREES Extension  procedures when      */
/*   this package is loaded.                                                  */
/*                                                                            */
/*----------------------------------------------------------------------------*/
int Sadie_TREES_Ext_Init(Tcl_Interp *interp) {
    
  Tcl_CreateCommand(interp, "Sadie_TREES_Ext_TrackRings", Sadie_TREES_Ext_TrackRingsCmd, 
        (ClientData) NULL, NULL);
  Tcl_CreateCommand(interp, "Sadie_TREES_Ext_SetUserInput", Sadie_TREES_Ext_SetUserInputCmd, 
        (ClientData) NULL, NULL);

  return TCL_OK;
}
 
