#ifndef BOUNDARY_H
#define BOUNDARY_H 1

#include "project_utilities.h"

BEGIN_C_DECLARATIONS

/**
 * Initial size of the extensible ring width array
 */
#define INIT_LEN 512

/**
 * Missing boundary angles and ring widths.
 * The values are outside the range possible for actual data. 
 */
#define NOT_AN_ANGLE -HUGE_VAL
#define NOT_A_WIDTH -HUGE_VAL

/**
 * Threshold for judging the equality of angles and distances.
 */
#define TINY (1e-16)

/**
 * Missing brightness profile value (which may also be a data value).
 */
#define MISSING_BRIGHTNESS DARKNESS

/**
 * Median of range of possible brightness values.
 * This assumes an 8-bit grayscale, and should really be an
 * image-specific variable.
 */
#define MID_GRAY ((PIXEL)127.0)

/**
 * Number of polynomial coefficients to derive temporary warped subimages.
 */
#define NCOEFF 4

/**
 * Number of control points used to derive temporary warped subimages.
 */
#define NPOINTS 4

/**
 * Maximum allowed difference between boundary angles from different methods.
 */
#define MAX_BOUND_ANGLE (0.45*M_PI)

/**
 * Angular interval within which to search for points on a boundary.
 */
#define SEARCH_TOLERANCE 0.35

/**
 * Tolerance for accepting the angles associated with points as equal.
 */
#define MATCH_TOLERANCE 0.125

/**
 * Match the angles associated with points to the line connecting them.
 */
#define MEASURE_TOLERANCE 0.083333333

/**
 * Number of points defining local neighborhoods on boundaries. 
 */
#define SRADIUS 5
#define SDIAMETER (2*SRADIUS+1)

/**
 * Fractional size of the margin at the edge of the warped subimage.
 */

#define SUBIMAGE_MARGIN_FRACTION 0.125

/* define round function */

/* #define round(x) ((x-floor(x))>0.5 ? (ceil(x)) : (floor(x))) */

/**
 * Accesses a point in a mosaic image through its index.
 * @param M  mosaic image data.
 * @param I  mosaic image index.
 * @param R  row (y) position.
 * @param C  column (x) position.
 */
#define mosaic_point(M,I,R,C) (M->data[0][R - I->voffset[C]][C])

/**
 * Projects a point onto the principal axis of a boundary.
 * @param P  the point to project.
 * @param A  the boundary with the axis to project it onto.
 * @param X  x displacement from boundary centroid.
 * @param Y  y displacement from boundary centroid.
 * @param Z  distance from centroid to projected position.
 */
#define project_on_principal_axis(P,A,X,Y,Z) X = P->xpos - A->centroid.xpos; \
  Y = P->ypos - A->centroid.ypos; \
  Z = Y*cos(A->centroid.theta) + X*sin(A->centroid.theta)

/**
 * Help re-balance a red-black tree.
 * Re-balancing involves two operations that are completely symmetrical,
 * except that the roles of left and right links are interchanged, so
 * this macro can generate the code for either operation.
 * Rebalancing depends on the parent of the parent of the node being
 * considerd, and the adjustment may be a simple re-coloring of nodes,
 * or it may involve changing their relative positions (with the
 * current node moving closer to the root), in which case we distinguish
 * two sub-cases, depending on whether the current node and its immediate
 * parent change places, before it changes levels.
 * @param U  either right or left.
 * @param D  either right or left (but differing from U).
 * @see rb_rebalance
 */
#define rb_balance(U,D) tmp = sub->parent->parent->D; \
if((tmp!=NULL) &&(tmp->rbcolor==red) ) \
  { \
    sub->parent->rbcolor = black; \
    tmp->rbcolor = black; \
    sub->parent->parent->rbcolor = red; \
    sub = sub->parent->parent; \
  } \
else \
  { \
    if (sub == sub->parent->D) \
      { \
        sub = sub->parent; \
        rb_rotate_##U (root,sub) ; \
      } \
    sub->parent->rbcolor = black; \
    sub->parent->parent->rbcolor = red; \
    rb_rotate_##D (root, sub->parent->parent) ; \
  } \

/**
 * Generates either of the two node-rotation functions for a red-black tree.
 * The two rotations are completely symmetrical, except that the roles
 * of right and left nodes are interchanged, so this macro is a code
 * template for generating either a right or a left rotation (note
 * that the function names are part of what's substituted).   A rotation
 * essentially operates on a group of three nodes, initially at three
 * different levels in the tree.  Two of the nodes move closer to the root,
 * one in the opposite direction, so we end up with just two levels after
 * the transformation.
 * @param U  either right or left.
 * @param D  either right or left (but differing from U).
 * @see rb_rebalance
 */
#define make_rotator(U,D) static void \
rb_rotate_##U (RBNODE**root, RBNODE*sub)  \
{ \
  RBNODE * tmp; \
  tmp = sub->D; \
  sub->D = tmp->U; \
  if(tmp->U != NULL)  \
    tmp->U->parent = sub; \
  tmp->parent = sub->parent; \
  if(sub == *root) \
    { \
      *root = tmp; \
    } \
  else if (sub == sub->parent->U) \
    { \
      sub->parent->U = tmp; \
    } \
  else \
    { \
      sub->parent->D = tmp; \
    } \
  tmp->U = sub; \
  sub->parent = tmp; \
} \

/**
 * A rectangle aligned with the x and y axes.
 */
typedef struct
{
  double minx; /**< minimum x coordinate. */
  double miny; /**< minimum y coordinate. */
  double maxx; /**< maximum x coordinate. */
  double maxy; /**< maximum y coordinate. */
} RECTANGLE;

/**
 * A point in a temporary temporary ring boundary representation.
 */
typedef struct boundpoint
{
  double xpos; /**< point x coordinate. */
  double ypos; /**< point y coordinate. */
  double theta; /**< local tangential angle. */
  int valid; /**< really Boolean: if directly determined or interpolated */
  struct boundpoint *next; /**< singly-linked lists represent boundaries */
} BOUNDPOINT;

typedef enum
{ red = 0, black = 1 } rbcolortype;

typedef struct rbnode
{
  rbcolortype rbcolor;
  double key;
  BOUNDPOINT *point;
  struct rbnode *parent;
  struct rbnode *left;
  struct rbnode *right;
} RBNODE;

typedef struct
{
  RECTANGLE enclosure;
  BOUNDPOINT centroid;
  BOUNDPOINT *points;
  BOUNDPOINT *endpoint;
  RBNODE *point_index;
  int stale;
} BOUNDER;

typedef struct
{
  double *d;
  int n;
  int block;
} WIDTHS;

typedef struct
{
  long double xsum;
  long double xsumsq;
  long double ysum;
  long double ysumsq;
  long double prodsum;
  int n;
} SUMSTATE;

typedef struct imagecontext
{
  void *image;
  void *imageindex;
  double (*accessor) (double, double, struct imagecontext *);
} IMAGE_CONTEXT;

#define image_access(X,Y,I) (I->accessor(X, Y, I))

typedef struct boundarycontext
{
  void *points;
  int hasnext;
  int prevx;
  int prevy;
  double xpos;
  double ypos;
  int valid;
  int nterms;
  double *xc;
  double *yc;
  struct boundarycontext *(*iterator) (struct boundarycontext *);
} BOUNDARY_CONTEXT;

#define iterate(B) (B = B->iterator(B))

double find_width (BOUNDER * in, BOUNDER * out, IMAGE_CONTEXT * mosaic,
                   PROFILE_LIST * last_ringprofile);

BOUNDER * make_boundary (BOUNDARY_CONTEXT * bcon);

void discard_boundary (BOUNDER * bound);

void initialize_mosaic_context (IMAGE_BYTE * im, MOSAIC_INDEX * imindex,
                                IMAGE_CONTEXT * newcontext);

void initialize_chaincode_context (LIST_NODE * chain_boundary,
                                   BOUNDARY_CONTEXT * newcontext);

void initialize_image_context (IMAGE * im, IMAGE_CONTEXT * newcontext);

void build_index (BOUNDER * bound);

void demolish_index (RBNODE * ix);

END_C_DECLARATIONS

#endif /* !BOUNDARY_H */
