/* General support for chain code boundary representations
 *
 * $Id: chain_code.c,v 1.7 2004/11/12 20:14:02 mmunro Exp $
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include  <stdio.h>
#include  <sadie.h>
#include  "mosaic.h"
#include  "chain.h"
#include  "logerr.h"

/**
 * Delete a chain-coded boundary list, freeing all allocated storage.
 * We explicitly reset the pointer to the freed boundaries to NULL,
 * rather than leaving it pointing to the invalid (freed) list head;
 * all the structures must have been dynamically allocated.
 * @param  thelisthandle  points to a pointer to the head of the list.
 */
void
demolish_chainlist (LIST_NODE ** thelisthandle)
{
  LIST_NODE *bnode = NULL;
  LIST_NODE *bnode_prev = NULL;
  CHAIN_NODE *cnode = NULL;
  CHAIN_NODE *cnode_prev = NULL;

  if (thelisthandle != NULL)
    {
      bnode = *thelisthandle;
      while (bnode != NULL)
        {
          bnode_prev = bnode;
          bnode = bnode->next;
          cnode = bnode_prev->chain;
          while (cnode != NULL)
            {
              cnode_prev = cnode;
              cnode = cnode->next;
              free (cnode_prev);
            }
          free(bnode_prev);
        }
      *thelisthandle = NULL;
    }
}

/**
 * Delete a chain map structure, freeing all dynamically allocated storage.
 * We explicitly reset the pointer to the freed structure to NULL,
 * rather than leaving it pointing to something invalid.  The chainmap must
 * have been allocated by malloc or something equivalent.
 * @param thechainmaphandle  points to a pointer to the chainmap.
 */
void
demolish_chainmap (CHAIN ** thechainmaphandle)
{
  if ((thechainmaphandle != NULL) && (*thechainmaphandle != NULL))
    {
      demolish_chainlist (&((*thechainmaphandle)->list));
      free (*thechainmaphandle);
      *thechainmaphandle = NULL;
    }
}

/**
 * Make a deep copy of just the head of a list of chain-coded boundaries.
 * @param oldhead  points to the head of the list to copy.
 * @param newhandle  points to where to store a pointer to the copy. 
 */
void
copy_chainlist_head (LIST_NODE * oldhead, LIST_NODE ** newhandle)
{
  CHAIN_NODE *oldchain = NULL;
  CHAIN_NODE **c_handle = NULL;

  logerr_assert (LOGERR_CRIT, LOGERR_FUNC, (newhandle != NULL));
  if (oldhead == NULL)
    *newhandle = NULL;
  else
    {
      *newhandle = malloc (sizeof (LIST_NODE));
      logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (*newhandle != NULL));
      **newhandle = *oldhead;
      (*newhandle)->next = NULL;
      c_handle = &((*newhandle)->chain);
      oldchain = oldhead->chain;
      while (oldchain != NULL)
        {
          *c_handle = malloc (sizeof (CHAIN_NODE));
          logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (*c_handle != NULL));
          **c_handle = *oldchain;
          c_handle = &((*c_handle)->next);
          oldchain = oldchain->next;
        }
      logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (*c_handle == NULL));
    }
}

/**
 * Make a deep copy of a list of chain-coded boundaries.
 * @param  oldlist  points to the head of the list to copy.
 * @return  a pointer to the head of the newly allocated copy of the list.
 */
LIST_NODE *
copy_chainlist (LIST_NODE * oldlist)
{
  LIST_NODE *newlist = NULL;
  LIST_NODE **b_handle = NULL;

  b_handle = &newlist;
  while (oldlist != NULL)
    {
      copy_chainlist_head (oldlist, b_handle);
      logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (*b_handle != NULL));
      b_handle = &((*b_handle)->next);
    }
  logerr_assert (LOGERR_DEBUG, LOGERR_FUNC, (*b_handle == NULL));
  return newlist;
}

/**
 * Find an item in a list of chain-coded boundaries by its id number.
 * We do a simple linear search for matching id numbers, stopping at the
 * first to be found, also finding its relative index position within
 * the list (counting from zero) and its immediate predecessor.  If we
 * cannot find a matching id number, the boundary results are NULL
 * and the index position is the list length.
 * @param id  the id number sought.
 * @param thelist  where to search for matching boundaries.
 * @param pred_handle  points to where we record the predecessor of the match.
 * @param match_handle  points to the where we record the matching item.
 * @param index  set to the index position of the matching item in the list.
 */
void
search_chainlist_by_id (int id, LIST_NODE * thelist, LIST_NODE ** pred_handle,
                        LIST_NODE ** match_handle, int * indexpos)
{
  logerr_assert (LOGERR_CRIT, LOGERR_FUNC,
                 (pred_handle && match_handle && indexpos));
  *indexpos = 0;
  *match_handle = thelist;
  while ((*match_handle != NULL) && ((*match_handle)->ring_id != id))
    {
      *pred_handle = *match_handle;
      *match_handle = (*match_handle)->next;
      (*indexpos)++;
    }
  if (*match_handle == NULL)
    *pred_handle = NULL;
}

/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function frees the memory block allocated for storing a chainmap.   */
/*                                                                            */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void RELMEM_CHAIN (
CHAIN *chainmap      /*  I   Pointer to the chain map.                        */
/*----------------------------------------------------------------------------*/
) {
    
  CHAIN *temp=NULL;
  LIST_NODE *templist=NULL, *curlist=NULL;
  CHAIN_NODE *tempnode=NULL, *curnode=NULL;
  if(chainmap) {
    curlist=chainmap->list;
    while(curlist!=NULL) {
      templist=curlist;
      curlist=curlist->next;
      curnode=templist->chain;
      while(curnode!=NULL) {
	tempnode=curnode;
	curnode=curnode->next;
	free(tempnode);
      }
      free(templist);
    }
    free(chainmap);
  }

  chainmap = NULL;
  return;
}


/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function computes the co-ordinates of a pixel in a chain given      */
/*   its chain code and co-ordinates of the preceeding pixel.                 */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void CHAIN_COMPUTE_COORDINATES(
int prevx,          /*  I   Column coordinate of previous pixel.              */
int prevy,          /*  I   Row coordinate of previous pixel.                 */
unsigned char code, /*  I   Chain code for the current pixel.                 */
int *currx,         /*  O   Pointer to the column coordinate of current pixel.*/   
int *curry          /*  O   Pointer to the row coordinate of current pixel.   */
/*----------------------------------------------------------------------------*/
) {

    
    switch (code) {
    case 0 :
	*curry = prevy - 1;
	*currx = prevx;
	break;
	
    case 1 :
	*curry = prevy - 1;
	*currx = prevx + 1;
	break;
	
    case 2 :
	*curry = prevy;
	*currx = prevx + 1;
	break;
	
    case 3 :
	*curry = prevy + 1;
	*currx = prevx + 1;
	break;
	
    case 4 :
	*curry = prevy + 1;
	*currx = prevx;
	break;
	
    case 5 :
	*curry = prevy + 1;
	*currx = prevx - 1;
	break;
	
    case 6 :
	*curry = prevy;
	*currx = prevx - 1;
	break;
	
    case 7 :
	*curry = prevy - 1;
	*currx = prevx - 1;
	break;

    default:
      logerr_debug (LOGERR_FUNC, "Meaningless chain code.");
	*curry = prevy;
	*currx = prevx;
    }

    return;
}


/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function computes the chain code of a pixel in a chain given        */
/*   its co-ordinates and that of the preceeding pixel.                       */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void CHAIN_COMPUTE_CODE (
int currx,           /*  I   Row coordinate of the current pixel.             */
int curry,           /*  I   Column coordinate of the current pixel.          */
int x,               /*  I   Row coordinate of the next pixel.                */
int y,               /*  I   Column coordinate of the next pixel.             */
unsigned char *code  /*  O   Pointer to thechain code.                        */
/*----------------------------------------------------------------------------*/
) {

    if (x == currx - 1 && y == curry - 1) {
	*code = 7;
    }
    else if (x == currx - 1 && y == curry) {
	*code = 0;
    }
    else if (x == currx - 1 && y == curry + 1) {
	*code = 1;
    }
    else if (x == currx && y == curry - 1) {
	*code = 6;
    }
    else if (x == currx && y == curry + 1) {
	*code = 2;
    }
    else if (x == currx + 1 && y == curry - 1) {
	*code = 5;
    }
    else if (x == currx + 1 && y == curry) {
	*code = 4;
    }
    else if (x == currx + 1 && y == curry + 1) {
	*code = 3;
   }

    return;
}
