/** @file chain_ringwidths.c
 * Measures ring widths and brightness profiles.
 * We're assuming that all the previous phases of TREES are complete:
 * that it has acquired images, assembled them into a mosaic, and
 * identified the ring boundaries.  This has one interface
 * visible to the rest of TREES, the function CHAIN_RINGWIDTHS,
 * 
 * @author University of Arizona Digital Image Analysis Lab
 * @date 2003
 * @version $Id: chain_ringwidths.c,v 1.28 2004/11/09 08:01:32 mmunro Exp $
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <math.h>
#include "junkmath.h"
#include "sadie_byte.h"
#include "chain.h"
#include "mosaic.h"
#include "trees_profile.h"
#include "boundary.h"
#if WITH_DMALLOC
#include <dmalloc.h>
#endif /* WITH_DMALLOC */
/**
 * Stopgap error handler.
 * This doesn't just abort the functions defined in this file,
 * or the thread running them, but the entire process running TREES.
 */
#define crashandburn(e) {fprintf(stderr,"Panic: %s!\n",e) ;exit(-2) ;} \

static void
initialize_sum_state (SUMSTATE * newstate, double x, double y)
{
  newstate->xsum = x;
  newstate->xsumsq = x * x;
  newstate->ysum = y;
  newstate->ysumsq = y * y;
  newstate->prodsum = x * y;
  newstate->n = 1;
}

static void
update_sum_state (SUMSTATE * thestate, double x, double y)
{
  thestate->xsum += x;
  thestate->xsumsq += x * x;
  thestate->ysum += y;
  thestate->ysumsq += y * y;
  thestate->prodsum += x * y;
  thestate->n++;
}

static void
downdate_sum_state (SUMSTATE * thestate, double x, double y)
{
  thestate->xsum -= x;
  thestate->xsumsq -= x * x;
  thestate->ysum -= y;
  thestate->ysumsq -= y * y;
  thestate->prodsum -= x * y;
  thestate->n--;
}

static double
axis_from_sum (SUMSTATE * sm)
{
  long double a, b, c;
  a = sm->ysumsq - (sm->ysum * sm->ysum) / sm->n;
  c = sm->xsumsq - (sm->xsum * sm->xsum) / sm->n;
  b = 2 * (sm->prodsum - (sm->xsum * sm->ysum) / sm->n);
  // if (b < TINY || (fabsl(a - c) < TINY)) /* commented */
  //  return 0.0;
  //else
    return (double) (atan2l (b, a - c) / 2.0);
}

static double
xmean_from_sum (SUMSTATE * sm)
{
  return (double) (sm->xsum / sm->n);
}

static double
ymean_from_sum (SUMSTATE * sm)
{
  return (double) (sm->ysum / sm->n);
}

static BRIGHTNESS
mosaic_image_accessor (double x, double y, IMAGE_CONTEXT * thiscontext)
{
  double x0, y0, gl1, gl2;
  int ix0, ix1, iy0, iy1;
  IMAGE_BYTE *im;
  MOSAIC_INDEX *imindex;

  /*<<<<<<< chain_ringwidths.c*/
  /*y = -y;*/
  im = (IMAGE_BYTE *) thiscontext->image;
  imindex = (MOSAIC_INDEX *) thiscontext->imageindex;
  x0 = floor (x);
  y0 = floor (y);
  ix0 = (int) x0;
  ix1 = ix0 + 1;
  iy0 = (int) y0;
  iy1 = iy0 + 1;
  if ((ix0 >= imindex->hstart) && (ix0 <= imindex->hend)
      && (iy0 >= imindex->vstart[ix0]) && (iy0 <= imindex->vend[ix0])
      && (ix1 >= imindex->hstart) && (ix1 <= imindex->hend)
      && (iy1 >= imindex->vstart[ix1]) && (iy1 <= imindex->vend[ix1]))
    {
      gl1 =
        (double) mosaic_point (im, imindex, iy0, ix0)
        + (x - x0) * (double) (mosaic_point (im, imindex, iy1, ix0)
                               - mosaic_point (im, imindex, iy0, ix0));
      gl2 =
        (double) mosaic_point (im, imindex, iy1, ix0)
        + (x - x0) * (double) (mosaic_point (im, imindex, iy1, ix1)
                               - mosaic_point (im, imindex, iy1, ix0));
      return (BRIGHTNESS) (gl1 + (y - y0) * (gl2 - gl1));
    }
  else
    return MISSING_BRIGHTNESS;
}

void
initialize_mosaic_context (IMAGE_BYTE * im, MOSAIC_INDEX * imindex,
                           IMAGE_CONTEXT * newcontext)
{
  newcontext->image = (void *) im;
  newcontext->imageindex = (void *) imindex;
  newcontext->accessor = mosaic_image_accessor;
}

static BRIGHTNESS
image_accessor (double x, double y, IMAGE_CONTEXT * thiscontext)
{
  double x0, y0, gl1, gl2;
  int ix0, ix1, iy0, iy1;
  IMAGE *im;

  im = (IMAGE *) thiscontext->image;
  x0 = floor (x);
  y0 = floor (y);
  ix0 = (int) x0;
  ix1 = ix0 + 1;
  iy0 = (int) y0;
  iy1 = iy0 + 1;
  if ((ix0 >= 0) && (iy0 >= 0) && (ix1 < im->npix) && (iy1 < im->nlin))
    {
      gl1 =
        (double) im->data[0][iy0][ix0]
        + (x - x0) * (double) (im->data[0][iy1][ix0] - im->data[0][iy0][ix0]);
      gl2 =
        (double) im->data[0][iy1][ix0]
        + (x - x0) * (double) (im->data[0][iy1][ix1] - im->data[0][iy1][ix0]);
      return (BRIGHTNESS) (gl1 + (y - y0) * (gl2 - gl1));
    }
  else
    return MISSING_BRIGHTNESS;
}

void
initialize_image_context (IMAGE * im, IMAGE_CONTEXT * newcontext)
{
  newcontext->image = (void *) im;
  newcontext->imageindex = NULL;
  newcontext->accessor = image_accessor;
}

static BOUNDARY_CONTEXT *
chaincode_iterator (BOUNDARY_CONTEXT * thiscontext)
{
  int newx, newy;
  CHAIN_NODE *cnode;

  if (!thiscontext->hasnext)
    crashandburn ("iterating off the end of chaincode list");
  thiscontext->hasnext = (thiscontext->points != NULL);
  if (thiscontext->hasnext)
    {
      cnode = (CHAIN_NODE *) thiscontext->points;
      CHAIN_COMPUTE_COORDINATES (thiscontext->prevx, thiscontext->prevy,
                                 cnode->code, &newx, &newy);
      thiscontext->xpos = (double) newx;
      thiscontext->ypos = (double) newy;
      thiscontext->prevx = newx;
      thiscontext->prevy = newy;
      thiscontext->points = (void *) cnode->next;
    }
  return thiscontext;
}

void
initialize_chaincode_context (LIST_NODE * chain_boundary,
                              BOUNDARY_CONTEXT * newcontext)
{ 
  newcontext->hasnext = (chain_boundary != NULL);
  if (newcontext->hasnext)
    {
      newcontext->points = (void *) chain_boundary->chain;
      newcontext->prevx = chain_boundary->startk;
      newcontext->prevy = chain_boundary->startj;
      newcontext->xpos = (double) newcontext->prevx;
      newcontext->ypos = (double) newcontext->prevy;
      newcontext->valid = ((chain_boundary->ring_type == NORMAL)
                           || (chain_boundary->ring_type == ADDED));
      newcontext->nterms = 0;
      newcontext->xc = NULL;
      newcontext->yc = NULL;
    }
  newcontext->iterator = chaincode_iterator;
}

static BOUNDARY_CONTEXT *
boundpoint_iterator (BOUNDARY_CONTEXT * thiscontext)
{
  BOUNDPOINT *pt;
  double x, y, xp, yp;

  if (!thiscontext->hasnext)
    crashandburn ("iterating off the end of boundary point list");
  thiscontext->hasnext = (thiscontext->points != NULL);
  if (thiscontext->hasnext)
    {
      pt = (BOUNDPOINT *) thiscontext->points;
      xp = pt->xpos;
      yp = pt->ypos;
      x = y = 0.0;
      switch (thiscontext->nterms)
        {
        case 6:
          x += thiscontext->xc[5] * yp * yp;
          y += thiscontext->yc[5] * yp * yp;
        case 5:
          x += thiscontext->xc[4] * xp * xp;
          y += thiscontext->yc[4] * xp * xp;
        case 4:
          x += thiscontext->xc[3] * xp * yp;
          y += thiscontext->yc[3] * xp * yp;
        case 3:
          x +=
            thiscontext->xc[0] +
            thiscontext->xc[1] * xp + thiscontext->xc[2] * yp;
          y +=
            thiscontext->yc[0] +
            thiscontext->yc[1] * xp + thiscontext->yc[2] * yp;
        }
      thiscontext->xpos = x;
      thiscontext->ypos = y;
      thiscontext->valid = pt->valid;
      thiscontext->points = (void *) pt->next;
    }
  return thiscontext;
}

static void
initialize_boundpoint_context (BOUNDPOINT * point_boundary,
                               BOUNDARY_CONTEXT * newcontext, int ncoeff,
                               double *xcoeff, double *ycoeff)
{
  newcontext->hasnext = (point_boundary != NULL);
  if (newcontext->hasnext)
    {
      newcontext->points = (void *) point_boundary->next;
      newcontext->prevx = 0;
      newcontext->prevy = 0;
      newcontext->xpos = point_boundary->xpos;
      newcontext->ypos = point_boundary->ypos;
      newcontext->valid = point_boundary->valid;
      newcontext->nterms = ncoeff;
      newcontext->xc = xcoeff;
      newcontext->yc = ycoeff;
    }
  newcontext->iterator = boundpoint_iterator;
}

static void
add_localprofile (BOUNDPOINT * start, BOUNDPOINT * finish, double locallength,
                  IMAGE_CONTEXT * img, PROFILE_LIST * last_localprofile)
{
  PROFILE_LIST localprofile;
  int i, nsteps;
  double deltax, deltay, x, y, scale;

  nsteps = (int) round (locallength);
  if (nsteps <= 0)
    return;
  localprofile = new_profile (nsteps);
  deltax = finish->xpos - start->xpos;
  deltay = finish->ypos - start->ypos;
  scale = 0;
  for (i = 0; i < nsteps; i++)
    {
      scale = (double) i / (double) nsteps;
      x = start->xpos + deltax * scale;
      y = start->ypos + deltay * scale;
      localprofile->val[i] = image_access (x, y, img);
    }
  (*last_localprofile)->next = localprofile;
  *last_localprofile = localprofile;
}

static void
demolish_points (BOUNDPOINT * victim)
{
  BOUNDPOINT *successor;
  while (victim != NULL)
    {
      successor = victim->next;
      free (victim);
      victim = successor;
    }
}

make_rotator (left, right)

make_rotator (right, left)

static void
rb_rebalance (RBNODE ** root, RBNODE * sub)
{
  RBNODE *tmp;
  sub->rbcolor = red;
  while ((sub != *root) && (sub->parent->rbcolor == red))
    {
      if (sub->parent == sub->parent->parent->left)
        {
          rb_balance (left, right);
        }
      else
        {
          rb_balance (right, left);
        }
    }
  (*root)->rbcolor = black;
}

static void
rb_insert (RBNODE ** root, BOUNDPOINT * bpt, double keyval)
{
  RBNODE *s;
  RBNODE *prev;
  RBNODE *newentry;
  if ((newentry = (RBNODE *) malloc (sizeof (RBNODE))) == NULL)
    crashandburn ("allocating memory for new point index record");
  newentry->left = NULL;
  newentry->right = NULL;
  newentry->point = bpt;
  newentry->key = keyval;
  if (*root == NULL)
    {
      newentry->rbcolor = black;
      newentry->parent = NULL;
      *root = newentry;
    }
  else
    {
      s = *root;
      do
        {
          prev = s;
          s = (keyval < s->key) ? s->left : s->right;
        }
      while (s != NULL);
      if (keyval < prev->key)
        prev->left = newentry;
      else
        prev->right = newentry;
      newentry->parent = prev;
      rb_rebalance (root, newentry);
    }
}

void
demolish_index (RBNODE * ix)
{
  if (ix == NULL)
    return;
  demolish_index (ix->right);
  demolish_index (ix->left);
  free (ix);
}

void
discard_boundary (BOUNDER * bound)
{
  if (bound->points != NULL)
    demolish_points (bound->points);
  if (bound->point_index != NULL)
    demolish_index (bound->point_index);
  free (bound);
}

static BOUNDPOINT *
transcribe_point (BOUNDPOINT * current, BOUNDARY_CONTEXT * bcon,
                  SUMSTATE * loc, SUMSTATE * glob, RECTANGLE * limits)
{
  BOUNDPOINT *newpoint;
  if ((newpoint = (BOUNDPOINT *) malloc (sizeof (BOUNDPOINT))) == NULL)
    crashandburn ("allocating memory for new boundary point");
  newpoint->xpos = bcon->xpos;
  if (bcon->xpos < limits->minx)
    limits->minx = bcon->xpos;
  if (bcon->xpos > limits->maxx)
    limits->maxx = bcon->xpos;
  newpoint->ypos = bcon->ypos;
  if (bcon->ypos < limits->miny)
    limits->miny = bcon->ypos;
  if (bcon->ypos > limits->maxy)
    limits->maxy = bcon->ypos;
  newpoint->theta = NOT_AN_ANGLE;
  newpoint->valid = bcon->valid;
  newpoint->next = NULL;
  current->next = newpoint;
  update_sum_state (loc, newpoint->xpos, newpoint->ypos);
  update_sum_state (glob, newpoint->xpos, newpoint->ypos);
  return newpoint;
}

BOUNDER *
make_boundary (BOUNDARY_CONTEXT * bcon)
{
  int len;
  BOUNDER *newrec;
  BOUNDPOINT *pointlist;
  BOUNDPOINT *last;
  BOUNDPOINT *seg_base;
  BOUNDPOINT *forget;
  SUMSTATE local, whole;
  RECTANGLE limits;
  if (!bcon->hasnext)
    crashandburn ("boundary contained no points");

  if ((pointlist = (BOUNDPOINT *) malloc (sizeof (BOUNDPOINT))) == NULL)
    crashandburn ("allocating memory for first boundary point");
  pointlist->xpos = bcon->xpos;
  limits.minx = limits.maxx = bcon->xpos;
  pointlist->ypos = bcon->ypos;
  limits.miny = limits.maxy = bcon->ypos;
  pointlist->theta = NOT_AN_ANGLE;
  pointlist->next = NULL;
  pointlist->valid = bcon->valid;
  initialize_sum_state (&whole, bcon->xpos, bcon->ypos);
  initialize_sum_state (&local, bcon->xpos, bcon->ypos);
  last = forget = seg_base = pointlist;
  iterate (bcon);

  for (len = 1; (bcon->hasnext) && (len <= SRADIUS); len++)
    {
      last = transcribe_point (last, bcon, &local, &whole, &limits);
      iterate (bcon);
    }

  while ((bcon->hasnext) && (len <= SDIAMETER))
    {
      last = transcribe_point (last, bcon, &local, &whole, &limits);
      seg_base->theta = axis_from_sum (&local);
      seg_base = seg_base->next;
      iterate (bcon);
      len++;
    }

  while (bcon->hasnext)
    {
      downdate_sum_state (&local, forget->xpos, forget->ypos);
      forget = forget->next;
      last = transcribe_point (last, bcon, &local, &whole, &limits);
      seg_base->theta = axis_from_sum (&local);
      seg_base = seg_base->next;
      iterate (bcon);
      len++;
    }
  len--;

  if (len >= SDIAMETER)
    {

      while (seg_base != NULL)
        {
          downdate_sum_state (&local, forget->xpos, forget->ypos);
          forget = forget->next;
          seg_base->theta = axis_from_sum (&local);
          seg_base = seg_base->next;
        }

    }

  if ((newrec = (BOUNDER *) malloc (sizeof (BOUNDER))) == NULL)
    crashandburn ("allocating memory for new boundary record");
  if (len <= 2)
    newrec->centroid.theta = NOT_AN_ANGLE;
  else
    newrec->centroid.theta = axis_from_sum (&whole);
  newrec->centroid.xpos = xmean_from_sum (&whole);
  newrec->centroid.ypos = ymean_from_sum (&whole);
  newrec->enclosure = limits;
  newrec->point_index = NULL;
  newrec->stale = 1;

  if (len < SDIAMETER)
    {
      demolish_points (pointlist);
      newrec->points = NULL;
      newrec->endpoint = NULL;
    }
  else
    {
      newrec->points = pointlist;
      newrec->endpoint = last;
    }
  return newrec;
}

void
build_index (BOUNDER * bound)
{
  BOUNDPOINT *curr;
  double dx, yp, dy;
  if ((bound == NULL) || (bound->points == NULL))
    return;
  bound->point_index = NULL;
  curr = bound->points;
  while (curr != NULL)
    {
      if (curr->valid)
        {
          project_on_principal_axis (curr, bound, dx, dy, yp);
          rb_insert (&bound->point_index, curr, yp);
        }
      curr = curr->next;
    }
}

static void
initialize_widths (WIDTHS * w)
{
  w->n = 0;
  if ((w->d = (double *) malloc (INIT_LEN * sizeof (double))) == NULL)
    crashandburn ("allocating memory for width measurements");
  w->block = INIT_LEN;
}

static void
push_width (double next_width, WIDTHS * w)
{
  double *newvec;
  if (w->n == w->block)
    {
      if ((newvec =
           (double *) malloc (2 * w->block * sizeof (double))) == NULL)
        crashandburn ("extending memory for width measurements");
      memcpy ((void *) newvec, (const void *) w->d,
              w->block * sizeof (double));
      free (w->d);
      w->d = newvec;
      w->block *= 2;
    }
  w->d[w->n++] = next_width;
}

static void
point_search (BOUNDPOINT * pt,
              double lob, double hib, RBNODE * sub,
              long double *widthsum, int *nwidths,
              IMAGE_CONTEXT * img, PROFILE_LIST * last_localprofile)
{
  double deltax, deltay, locallength;
  double thetamean, psi;
  if ((sub == NULL) || (lob > hib))
    return;
  if ((lob <= sub->key) && (sub->key <= hib)
      && (fabs (sub->point->theta - pt->theta) < MATCH_TOLERANCE))
    {
      deltax = sub->point->xpos - pt->xpos;
      deltay = pt->ypos - sub->point->ypos;
      thetamean = (sub->point->theta + pt->theta) / 2;
      psi = atan2 (deltay, deltax);
      if (fabs(thetamean - psi) < MEASURE_TOLERANCE)
        {
          locallength = sqrt (deltax * deltax + deltay * deltay);
          (*widthsum) += locallength;
          (*nwidths)++;
          add_localprofile (pt, sub->point, locallength, img,
                            last_localprofile);
        }
    }
  if (sub->key >= lob)
    point_search (pt, lob, fmin (hib, sub->key), sub->left, widthsum,
                  nwidths, img, last_localprofile);
  if (sub->key <= hib)
    point_search (pt, fmax (lob, sub->key), hib, sub->right, widthsum,
                  nwidths, img, last_localprofile);
}

static void
segment_dist (BOUNDPOINT * pt, BOUNDER * bound, long double *widthsum,
              int *nwidths, IMAGE_CONTEXT * img,
              PROFILE_LIST * last_localprofile)
{
  double xp, tol;
  double dx, yp, dy;
  project_on_principal_axis (pt, bound, dx, dy, yp);
  xp = dx * cos (bound->centroid.theta) - dy * sin (bound->centroid.theta);
  tol = fabs (xp * tan (SEARCH_TOLERANCE));
  point_search (pt, yp - tol, yp + tol, bound->point_index, widthsum,
                nwidths, img, last_localprofile);
}

static double
width_from_detail (BOUNDER * in, BOUNDER * out, IMAGE_CONTEXT * img,
                   PROFILE_LIST * last_localprofile)
{
  long double widthsum = 0.0;
  int nwidths = 0;
  BOUNDPOINT *current;
  if (out->stale)
    {
      build_index (out);
      out->stale = 0;
    }
  current = in->points;
  while (current != NULL)
    {
      if (current->valid)
        segment_dist (current, out, &widthsum, &nwidths, img,
                      last_localprofile);
      current = current->next;
    }
  if (nwidths > 0)
    return (double) (widthsum / nwidths);
  else
    return NOT_A_WIDTH;
}

static void
profiles_from_newimage (double width, BOUNDER * inner, BOUNDER * outer,
                        IMAGE_CONTEXT * img, PROFILE_LIST * last_localprofile)
{
  int xlo, ylo, xhi, yhi;
  int xmargin, ymargin;
  double dx, dy, dx0, dy0, dx1, dy1;
  double x[NPOINTS];
  double y[NPOINTS];
  double xabs[NPOINTS];
  double yabs[NPOINTS];
  double xp[NPOINTS];
  double yp[NPOINTS];
  double xcoeff[NCOEFF];
  double ycoeff[NCOEFF];
  double xc_inv[NCOEFF];
  double yc_inv[NCOEFF];
  BOUNDARY_CONTEXT inner_bcon, outer_bcon;
  BOUNDER *warpinner;
  BOUNDER *warpouter;
  IMAGE *temp_subimage;
  IMAGE *warped_image;
  IMAGE_CONTEXT warp;
  char fname [512];

  xlo = (int) floor (min (inner->enclosure.minx, outer->enclosure.minx));
  ylo = (int) floor (min (inner->enclosure.miny, outer->enclosure.miny));
  xhi = (int) ceil (max (inner->enclosure.maxx, outer->enclosure.maxx));
  yhi = (int) ceil (max (inner->enclosure.maxy, outer->enclosure.maxy));
  xmargin = (int) ((1 + xhi - xlo) * SUBIMAGE_MARGIN_FRACTION);
  ymargin = (int) ((1 + yhi - ylo) * SUBIMAGE_MARGIN_FRACTION);
  xlo -= xmargin;
  xhi += xmargin;
  ylo -= ymargin;
  yhi += xmargin;
  MOSAIC_BYTE_EXTRACT ((IMAGE_BYTE *) img->image,
                       (MOSAIC_INDEX *) img->imageindex, ylo, yhi, xlo, xhi,
                       MID_GRAY, &temp_subimage);
  xabs[0] = inner->points->xpos;
  yabs[0] = inner->points->ypos;
  xabs[1] = inner->endpoint->xpos;
  yabs[1] = inner->endpoint->ypos;
  xabs[2] = outer->points->xpos;
  yabs[2] = outer->points->ypos;
  xabs[3] = outer->endpoint->xpos;
  yabs[3] = outer->endpoint->ypos;
  x[0] = xabs[0] - xlo;
  y[0] = yabs[0] - ylo;
  x[1] = xabs[1] - xlo;
  y[1] = yabs[1] - ylo;
  x[2] = xabs[2] - xlo;
  y[2] = yabs[2] - ylo;
  x[3] = xabs[3] - xlo;
  y[3] = yabs[3] - ylo;
  dx0 = inner->endpoint->xpos - inner->points->xpos;
  dx1 = outer->endpoint->xpos - outer->points->xpos;
  dy0 = inner->endpoint->ypos - inner->points->ypos;
  dy1 = outer->endpoint->ypos - outer->points->ypos;
  xp[0] = xmargin;
  yp[0] = ymargin;
  xp[1] = xp[0];
  yp[1] = yp[0] + sqrt ((dx0 * dx0 + dx1 * dx1 + dy0 * dy0 + dy1 * dy1) / 2);
  xp[2] = xp[0] + width;
  yp[2] = yp[0];
  xp[3] = xp[2];
  yp[3] = yp[1];
  GEOMCOEF (x, y, xp, yp, NPOINTS, NCOEFF, xcoeff, ycoeff);
  GEOMCOEF (xp, yp, xabs, yabs, NPOINTS, NCOEFF, xc_inv, yc_inv);
  GEOMWARP (temp_subimage, 1, ceil (yp[1] + ymargin),
	    ceil (xp[2] + xmargin), xcoeff, ycoeff,
	    NCOEFF, 0.0, MID_GRAY, &warped_image);
  initialize_image_context (warped_image, &warp);
  initialize_boundpoint_context (inner->points, &inner_bcon,
                                 NCOEFF, xc_inv, yc_inv);
  initialize_boundpoint_context (outer->points, &outer_bcon,
                                 NCOEFF, xc_inv, yc_inv);
  warpinner = make_boundary (&inner_bcon);
  warpouter = make_boundary (&outer_bcon);
  if (width_from_detail (warpinner, warpouter, &warp, last_localprofile) ==
      NOT_A_WIDTH)
    {
      dx = warpouter->centroid.xpos - warpinner->centroid.xpos;
      dy = warpouter->centroid.ypos - warpinner->centroid.ypos;
      add_localprofile (&(warpinner->centroid), &(warpouter->centroid),
                        sqrt (dx * dx + dy * dy), &warp, last_localprofile);
    }
  discard_boundary (warpinner);
  discard_boundary (warpouter);
  if (CHECKIMG (temp_subimage))
    RELIMG (&temp_subimage);
  if (CHECKIMG (warped_image))
    RELIMG (&warped_image);
}

double
find_width (BOUNDER * in, BOUNDER * out, IMAGE_CONTEXT * mosaic,
            PROFILE_LIST * last_ringprofile)
{
  double width, deltax, deltay, checkw;
  PROFILE_LIST local_head, ring_mean;
  PROFILE_LIST last_localprofile;
  deltax = out->centroid.xpos - in->centroid.xpos;
  deltay = out->centroid.ypos - in->centroid.ypos;

  local_head = new_profile (0);
  last_localprofile = local_head;
  width = NOT_A_WIDTH;
  if ((out->centroid.theta != NOT_AN_ANGLE) && (in->points != NULL)
      && (out->points != NULL))
    {
      width = width_from_detail (in, out, mosaic, &last_localprofile);
      if ((width != NOT_A_WIDTH) && (in->centroid.theta != NOT_AN_ANGLE))
        {
          checkw =
            (fabs (deltay * sin (in->centroid.theta) + deltax * cos (in->centroid.theta)) +
             fabs (deltay * sin (out->centroid.theta) +
                   deltax * cos (out->centroid.theta))) / 2.0;
          if ((width < checkw / 2.0) || (width > checkw * 2.0))
            {
              width = sqrt (deltax * deltax + deltay * deltay);
            }
        }
    }
  if (width == NOT_A_WIDTH)
    {
      if ((out->centroid.theta == NOT_AN_ANGLE) && (in->centroid.theta == NOT_AN_ANGLE))
        width = sqrt (deltax * deltax + deltay * deltay);
      else if (out->centroid.theta == NOT_AN_ANGLE)
        width = fabs (deltay * sin (in->centroid.theta) + deltax * cos (in->centroid.theta));
      else if (in->centroid.theta == NOT_AN_ANGLE)
        width = fabs (deltay * sin (out->centroid.theta) + deltax * cos (out->centroid.theta));
      else if (fabs (out->centroid.theta - in->centroid.theta) < MAX_BOUND_ANGLE)
        {
          width =
            (fabs (deltay * sin (in->centroid.theta) + deltax * cos (in->centroid.theta)) +
             fabs (deltay * sin (out->centroid.theta) +
                   deltax * cos (out->centroid.theta))) / 2.0;
          profiles_from_newimage (width, in, out, mosaic, &last_localprofile);
        }
      else
        width = sqrt (deltax * deltax + deltay * deltay);
    }
  if (width == NOT_A_WIDTH)
    ring_mean = new_profile (0);
  else
    {
      ring_mean = new_profile ((int) round (width));
      find_mean_profile (ring_mean, local_head->next);
    }
  demolish_profiles (local_head);
  (*last_ringprofile)->next = ring_mean;
  *last_ringprofile = ring_mean;
  return width;
}

void
CHAIN_RINGWIDTHS (IMAGE_BYTE * mimage,
                  MOSAIC_INDEX * mindex,
                  CHAIN * chainmap,
                  PROFILE_LIST * profiles, int *numrings, double **ringwidths, BOUNDBOX_LIST * bound_box_list)
{

  LIST_NODE *chain_boundary;
  BOUNDER *inner = NULL;
  BOUNDER *outer = NULL;
  WIDTHS widthlist;
  PROFILE_LIST global_head;
  PROFILE_LIST last_ringprofile;
  IMAGE_CONTEXT mosaic;
  BOUNDARY_CONTEXT chaincontext;
  BOUNDBOX_LIST temp_boundbox=NULL;
  BOUNDBOX_LIST boundbox_head=NULL;

  initialize_widths (&widthlist);
  global_head = new_profile (0);
  if ((chainmap == NULL) || (chainmap->list == NULL))
    chain_boundary = NULL;
  else
    {
      chain_boundary = chainmap->list;
      while ((chain_boundary != NULL)
             && (chain_boundary->ring_type == REMOVED)) 
	chain_boundary = chain_boundary->next;
    }
  if (chain_boundary != NULL)
    {
      last_ringprofile = global_head;
      initialize_mosaic_context (mimage, mindex, &mosaic);
      initialize_chaincode_context (chain_boundary, &chaincontext);
      inner = make_boundary (&chaincontext);
      if ((temp_boundbox = (BOUNDBOX_LIST) malloc (sizeof (BOUNDBOX_NODE))) == NULL)
        crashandburn ("allocating memory for bounding boxes");
      temp_boundbox->minx = (inner->enclosure).minx;
      temp_boundbox->miny = (inner->enclosure).miny; 
      temp_boundbox->maxx = (inner->enclosure).maxx;
      temp_boundbox->maxy = (inner->enclosure).maxy;
      temp_boundbox->next = NULL;
      boundbox_head = temp_boundbox;
      do
        {
          chain_boundary = chain_boundary->next;
        }
      while ((chain_boundary != NULL)
             && (chain_boundary->ring_type == REMOVED));
    }      
  while (chain_boundary != NULL)
    {
      if ((temp_boundbox->next = (BOUNDBOX_LIST) malloc (sizeof (BOUNDBOX_NODE))) == NULL)
        crashandburn ("allocating memory for bounding boxes");
      temp_boundbox = temp_boundbox->next;
      initialize_chaincode_context (chain_boundary, &chaincontext);
      outer = make_boundary (&chaincontext);
      printf("\nProcessing ring id=%d\n", chain_boundary->ring_id);
      temp_boundbox->minx = (outer->enclosure).minx;
      temp_boundbox->miny = (outer->enclosure).miny; 
      temp_boundbox->maxx = (outer->enclosure).maxx;
      temp_boundbox->maxy = (outer->enclosure).maxy;
      temp_boundbox->next = NULL;
      push_width (find_width (inner, outer, &mosaic, &last_ringprofile),
                  &widthlist);
      discard_boundary (inner);
      inner = outer;
      do
        {
          chain_boundary = chain_boundary->next;
        }
      while ((chain_boundary != NULL)
             && (chain_boundary->ring_type == REMOVED));
    }
  if (inner)
    discard_boundary (inner);
  *bound_box_list = boundbox_head;
  *numrings = widthlist.n;
  *ringwidths = widthlist.d;
  *profiles = reduce_profiles (global_head);
}
