#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include "trees.h"
#include "histogram.h"



/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*  Description:   Function to initialize a histogram data structure.         */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void HIST_INIT (
HISTOGRAM  *hist   /* Pointer to the histogram data structure.                */
/*----------------------------------------------------------------------------*/
) {
    register int i;
    
    hist->bin = 0;
    hist->cbin = 0;
    hist->edge = 0;
    hist->numpixels = 0;
    hist->r = (double)0.0;
    hist->per = 0;
    for(i = 0; i < HIST_RESOL; i++)
	hist->dir[i] = hist->mag[i] = 0;
}



/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*  Description:   Function to compute the magnitude histogram.               */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void HIST_MAG (
IMAGE_BYTE  *iMag,        /*  I   Pointer to the gradient magnitude mosaic.   */
MOSAIC_INDEX  *index,     /*  I   Pointer to the mosaic index.                */
HISTOGRAM  *globalhist,   /*  I/O Pointer to the global histogram structure.  */
HISTOGRAM  **localhist,   /*  I/O Address of a pointer to the local histogram.*/
int nrows,                /*  I   Number of rows of blocks.                   */
int ncols                 /*  I   Number of cols of blocks.                   */
/*----------------------------------------------------------------------------*/
) {
    register int i, j, k;
    int row, col;

    int progmeter;
    double progress;
    double localpercent;
    double localprogress;
    int total, increment;
    int cancel=0;

    progmeter = ProgMeter_Create("Magnitude Histograms");
    progress = 0.0;

    if (progmeter) {
        localpercent = 0.75;
        localprogress = 0.0;
	total = index->hend - index->hstart;
        increment = rint((double)total/5.0) + 1;
    }

    /* Local */
    for(k = index->hstart; k <= index->hend; k++) {

	if (progmeter) {
	    if (((k-index->hstart)%increment) == 0) {
		localprogress = ((double)(k-index->hstart)/(double)total)*100.0;
		cancel = ProgMeter_Update(progmeter,(localprogress*localpercent)+progress);
	    }
	}

	for(j = index->vstart[k]; j <= index->vend[k]; j++) {
          row = (index->vstart[k]-index->roi->top[0])/HIST_BLOCK_DIM; /* Unresolved bug? */
	    col = (k-index->hstart)/HIST_BLOCK_DIM;
	    localhist[row][col].mag[iMag->data[0][j-index->voffset[k]][k]]++;
	    localhist[row][col].numpixels++; 
	}
    }
    
    if (progmeter) {
	progress = 75.0;
	cancel = ProgMeter_Update(progmeter,progress);
    }
    
    if (progmeter) {
        localpercent = 0.25;
        localprogress = 0.0;
	total = nrows;
        increment = rint((double)total/5.0) + 1;
    }

    /* Global */
    for(i = 0; i < nrows; i++) {
	if (progmeter) {
	    if ((i%increment) == 0) {
		localprogress = ((double)i/(double)total)*100.0;
		cancel = ProgMeter_Update(progmeter,(localprogress*localpercent)+progress);
	    }
	}

	for(j = 0; j < ncols; j++) {
	    for(k = 0; k < HIST_RESOL; k++) {
		globalhist->mag[k] += localhist[i][j].mag[k];
		globalhist->numpixels += localhist[i][j].numpixels;
	    }
	}
    }


    if (progmeter) {
	progress = 100.0;
	cancel = ProgMeter_Update(progmeter,progress);
    }

    if (progmeter) {
        ProgMeter_Destroy(progmeter);
    }
}



/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*  Description:   Function to compute the rth percentile.                    */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void HIST_RTH_PERCENTILE (
HISTOGRAM  *globalhist,   /*  I/O Pointer to the global histogram structure.  */
HISTOGRAM  **localhist,   /*  I/O Address of a pointer to the local histogram.*/
double r,                 /*  I   the percentile to compute.                  */
int nrows,                /*  I   Number of rows of blocks.                   */
int ncols                 /*  I   Number of cols of blocks.                   */
/*----------------------------------------------------------------------------*/
) {
    register int i, j, k;
    int sum;

    int progmeter;
    double progress;
    double localpercent;
    double localprogress;
    int total, increment;
    int cancel=0;

    progmeter = ProgMeter_Create("Computing Rth Percentile");
    progress = 0.0;

    if (progmeter) {
        localpercent = 1.0;
        localprogress = 0.0;
	total = nrows;
        increment = rint((double)total/5.0) + 1;
    }
    
    globalhist->r = r;
    
    /* Local */
    for(i = 0; i < nrows; i++) {

	if (progmeter) {
	    if ((i%increment) == 0) {
		localprogress = ((double)i/(double)total)*100.0;
		cancel = ProgMeter_Update(progmeter,(localprogress*localpercent)+progress);
	    }
	}

	for(j = 0; j < ncols; j++) {
	    localhist[i][j].r = r;
	    for(k = HIST_RESOL-1, sum = 0; k >= 0; k--) {
		sum += localhist[i][j].mag[k];
		if (sum >= localhist[i][j].numpixels*r) {
		    localhist[i][j].per = k;
		    break;
		}
	    }
	}
    }

    
    /* Global */
    for(k = HIST_RESOL-1, sum = 0; k >= 0; k--) {
	sum += globalhist->mag[k];
	if (sum >= globalhist->numpixels*r) {
	    globalhist->per = k;
	    break;
	}
    }

    if (progmeter) {
	progress = 100.0;
	cancel = ProgMeter_Update(progmeter,progress);
    }

    if (progmeter) {
        ProgMeter_Destroy(progmeter);
    }

}



/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*  Description:   Function to compute the direction histogram.               */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void HIST_DIR (
IMAGE_BYTE  *iMag,        /*  I   Pointer to the gradient magnitude mosaic.   */
IMAGE_BYTE  *iDir,        /*  I   Pointer to the gradient direction mosaic.   */
MOSAIC_INDEX  *index,     /*  I   Pointer to the mosaic index.                */
HISTOGRAM  *globalhist,   /*  I/O Pointer to the global histogram structure.  */
HISTOGRAM  **localhist,   /*  I/O Address of a pointer to the local histogram.*/
int nrows,                /*  I   Number of rows of blocks.                   */
int ncols                 /*  I   Number of cols of blocks.                   */
/*----------------------------------------------------------------------------*/
) {
    register int j, k;
    int row, col;

    int progmeter;
    double progress;
    double localpercent;
    double localprogress;
    int total, increment;
    int cancel=0;

    progmeter = ProgMeter_Create("Direction Histograms");
    progress = 0.0;

    if (progmeter) {
        localpercent = 1.0;
        localprogress = 0.0;
	total = index->hend - index->hstart;
        increment = rint((double)total/5.0) + 1;
    }
    
    for(k = index->hstart; k <= index->hend; k++) {

	if (progmeter) {
	    if (((k-index->hstart)%increment) == 0) {
		localprogress = ((double)(k-index->hstart)/(double)total)*100.0;
		cancel = ProgMeter_Update(progmeter,(localprogress*localpercent)+progress);
	    }
	}

	for(j = index->vstart[k]; j <= index->vend[k]; j++) {

	    row = (index->vstart[k]-index->roi->top[0])/HIST_BLOCK_DIM; /* Unresolved bug? */
	    col = (k-index->hstart)/HIST_BLOCK_DIM;

	    /* Local histogram */
	    if (iMag->data[0][j-index->voffset[k]][k] >= localhist[row][col].per)
		localhist[row][col].dir[iDir->data[0][j-index->voffset[k]][k]] += iMag->data[0][j-index->voffset[k]][k];

	    /* Global histogram */
	    if (iMag->data[0][j-index->voffset[k]][k] >= globalhist->per)
		globalhist->dir[iDir->data[0][j-index->voffset[k]][k]] += iMag->data[0][j-index->voffset[k]][k];
    
	}
    }

    if (progmeter) {
	progress = 100.0;
	cancel = ProgMeter_Update(progmeter,progress);
    }

    if (progmeter) {
        ProgMeter_Destroy(progmeter);
    }
}



/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*  Description:   Function to compute the orientation from the histogram.    */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void HIST_ORIENTATION (
HISTOGRAM  *globalhist,   /*  I/O Pointer to the global histogram structure.  */
HISTOGRAM  **localhist,   /*  I/O Address of a pointer to the local histogram.*/
int nrows,                /*  I   Number of rows of blocks.                   */
int ncols                 /*  I   Number of cols of blocks.                   */
/*----------------------------------------------------------------------------*/
) {
    register int i, j, k;
    int max, bin;
    
    int progmeter;
    double progress;
    double localpercent;
    double localprogress;
    int total, increment;
    int cancel=0;

    progmeter = ProgMeter_Create("Computing Ring Orientation");
    progress = 0.0;

    if (progmeter) {
        localpercent = 1.0;
        localprogress = 0.0;
	total = nrows;
        increment = rint((double)total/5.0) + 1;
    }
    
    /* Local */
    for(i = 0; i < nrows; i++) {

	if (progmeter) {
	    if ((i%increment) == 0) {
		localprogress = ((double)i/(double)total)*100.0;
		cancel = ProgMeter_Update(progmeter,(localprogress*localpercent)+progress);
	    }
	}

	for(j = 0; j < ncols; j++) {
	    for(k = 0, max = 0; k < HIST_RESOL; k++) {
		if (localhist[i][j].dir[k] > max) {
		    max = localhist[i][j].dir[k];
		    bin = k;
		}
	    }
	    localhist[i][j].bin = bin;
	    localhist[i][j].cbin = bin;
	}
    }

    /* Global */
    for(k = 0, max = 0; k < HIST_RESOL; k++) {
	if (globalhist->dir[k] > max) {
	    max = globalhist->dir[k];
	    bin = k;
	}
    }
    globalhist->bin = bin;
    globalhist->cbin = bin;

    if (progmeter) {
	progress = 100.0;
	cancel = ProgMeter_Update(progmeter,progress);
    }

    if (progmeter) {
        ProgMeter_Destroy(progmeter);
    }
    
}



/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*  Description:   Function to determine orientation of the block that        */
/*                 lies in the LOCAL_DEVN of global orientation.              */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
int HIST_DIR_OF_EDGE_NEIGHBOUR (
HISTOGRAM  *globalhist,   /*  I   Pointer to the global histogram structure.  */
HISTOGRAM  **localhist,   /*  I   Address of a pointer to the local histogram.*/
int nrows,                /*  I   Number of rows of blocks.                   */
int ncols,                /*  I   Number of cols of blocks.                   */
int i,                    /*  I   Row co-ordinate of block.                   */
int j                     /*  I   Column co-ordinate of block.                */
/*----------------------------------------------------------------------------*/
) {
    int k, l;
    
    for(k = i-1; k < i+2; k ++) {
	if (k < 0 || k >= nrows)
	    continue;
	for(l = j-1; l < j+2; l ++) {
	    if (l < 0 || l >= ncols)
		continue;
	    
	    if ( abs(localhist[k][l].bin - globalhist->bin) < LOCAL_DEVN )
		return(localhist[k][l].bin);
	}
    }

    return(-1);
}



/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*  Description:   Function to compute the orientation from the histogram.    */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void HIST_RESOLVE (
HISTOGRAM  *globalhist,   /*  I   Pointer to the global histogram structure.  */
HISTOGRAM  **localhist,   /*  I/O Address of a pointer to the local histogram.*/
int nrows,                /*  I   Number of rows of blocks.                   */
int ncols                 /*  I   Number of cols of blocks.                   */
/*----------------------------------------------------------------------------*/
) { 
    register int i, j;
    int bin;
    
    int progmeter;
    double progress;
    double localpercent;
    double localprogress;
    int total, increment;
    int cancel=0;

    progmeter = ProgMeter_Create("Resolving Ring Orientation");
    progress = 0.0;

    if (progmeter) {
        localpercent = 1.0;
        localprogress = 0.0;
	total = nrows;
        increment = rint((double)total/5.0) + 1;
    }

    for(i = 0; i < nrows; i++) {

	if (progmeter) {
	    if ((i%increment) == 0) {
		localprogress = ((double)i/(double)total)*100.0;
		cancel = ProgMeter_Update(progmeter,(localprogress*localpercent)+progress);
	    }
	}

	for(j = 0; j < ncols; j++) {
	    /* Check if the local gradient orientation lies outside the LOCAL_DEVN
	       neighbourhood of global orientation */
	    if ( abs(localhist[i][j].bin-globalhist->bin) > LOCAL_DEVN ) {
		
		/* Check to see if the local orientation of any of the neighbours
		   lies within a LOCAL_DEVN neighbourhood of global orientation */
		bin = HIST_DIR_OF_EDGE_NEIGHBOUR(globalhist, localhist, nrows, ncols, i, j);

		/* If the block does not have an edge neighbour
		   return value is -1 */
		if (bin == -1) {
		    localhist[i][j].bin = globalhist->bin;
		} else {
		    localhist[i][j].bin = bin;
		}
		
	    }
	}
   }

    if (progmeter) {
	progress = 100.0;
	cancel = ProgMeter_Update(progmeter,progress);
    }

    if (progmeter) {
        ProgMeter_Destroy(progmeter);
    }
    
}



/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*  Description:   Function to determine the number of ring neighbours.       */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
int HIST_NUM_RING_NEIGH(
IMAGE_BYTE  *iDir,        /*  I   Pointer to the gradient direction mosaic.   */
MOSAIC_INDEX  *index,     /*  I   Pointer to the mosaic index.                */
int m,                    /*  I   Row co-ordinate of pixel.                   */
int n,                    /*  I   Column co-ordinate of pixel.                */
int bin                   /*  I   Orientation of local block.                 */
/*----------------------------------------------------------------------------*/
) {
    int j, k;
    int diff, count = 0;

    for(j = m - 1; j < m + 2; j++) {
	for(k = n - 1; k < n + 2; k++) {
	    if (k < index->hstart || k > index->hend)
		continue;
	    if (j < index->vstart[k] || j > index->vend[k])
		continue;
	    diff = abs(iDir->data[0][j-index->voffset[k]][k] - bin);
	    if (diff < DIR_TOL || (255-diff) < DIR_TOL)
		count++;
	}
    }
    
    return(count);
}



/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*  Description:   Function to suppress the non-edge pixels.                  */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void HIST_SUPPRESS (
IMAGE_BYTE  *iMag,        /*  I/O Pointer to the gradient magnitude mosaic.   */
IMAGE_BYTE  *iDir,        /*  I   Pointer to the gradient direction mosaic.   */
MOSAIC_INDEX  *index,     /*  I   Pointer to the mosaic index.                */
HISTOGRAM  **localhist    /*  I   Address of a pointer to the local histogram.*/
/*----------------------------------------------------------------------------*/
) {
    register int j, k;
    int bin, diff, row, col;

    int progmeter;
    double progress;
    double localpercent;
    double localprogress;
    int total, increment;
    int cancel=0;

    progmeter = ProgMeter_Create("Suppress Non-edge pixels");
    progress = 0.0;

    if (progmeter) {
        localpercent = 1.0;
        localprogress = 0.0;
	total = index->hend - index->hstart;
        increment = rint((double)total/5.0) + 1;
    }

    for(k = index->hstart; k <= index->hend; k++) {

	if (progmeter) {
	    if (((k-index->hstart)%increment) == 0) {
		localprogress = ((double)(k-index->hstart)/(double)total)*100.0;
		cancel = ProgMeter_Update(progmeter,(localprogress*localpercent)+progress);
	    }
	}

	for(j = index->vstart[k]; j <= index->vend[k]; j++) {
	    row = (index->vstart[k]-index->roi->top[0])/HIST_BLOCK_DIM; /* Unresolved bug? */
	    col = (k-index->hstart)/HIST_BLOCK_DIM;
	    bin = localhist[row][col].bin;
	    diff = abs(iDir->data[0][j-index->voffset[k]][k] - bin);

	    if (diff > DIR_TOL && (255-diff) > DIR_TOL) {
		if (HIST_NUM_RING_NEIGH(iDir, index, j, k, bin) < 2)
		    iMag->data[0][j-index->voffset[k]][k] = 0;
	    }
	}
    }

    if (progmeter) {
	progress = 100.0;
	cancel = ProgMeter_Update(progmeter,progress);
    }

    if (progmeter) {
        ProgMeter_Destroy(progmeter);
    }

}

/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*  Description:   Function to determine the neighbour with highest gradient. */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
int HIST_FIND_MAX_NEIGH (
IMAGE_BYTE  *iMag,        /*  I   Pointer to the gradient magnitude mosaic.   */
IMAGE_BYTE  *byteimg,     /*  I   Pointer to the mosaic image.                */
MOSAIC_INDEX  *index,     /*  I   Pointer to the mosaic index.                */
int startx,               /*  I   Row coordinate of pixel.                    */
int starty,               /*  I   Column coordinate of pixel.                 */
int *newx,                /*  I   Row coordinate of neighbour.                */
int *newy                 /*  I   Row coordinate of neighbour.                */
/*----------------------------------------------------------------------------*/
) {
    int i, j, maxx, maxy, max;
    
    if (startx + 1 == index->vend[starty])
	return(0);

    max = 0;
    for(j = starty - 1; j < starty + 2; j++) {
	
	if (j < index->hstart || j > index->hend)
	    continue;
	
	for(i = startx + 1; i < startx + 2; i++) {

	    if (i < index->vstart[j] || i > index->vend[j])
		continue;
	    
	    if (iMag->data[0][i - index->voffset[j]][j] > max) {
		max = iMag->data[0][i - index->voffset[j]][j];
		maxx = i;
		maxy = j;
	    }
	}
    }

    if (max) {
	*newx = maxx;
	*newy = maxy;
	return(1);
    }
    else
	return(0);

}


/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*  Description:   Function to determine the number of neighbours.            */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
int HIST_NUM_NEIGH(
IMAGE_BYTE  *iMag,        /*  I   Pointer to the gradient magnitude mosaic.   */
MOSAIC_INDEX  *index,     /*  I   Pointer to the mosaic index.                */
int m,                    /*  I   Row coordinate of pixel.                    */
int n                    /*  I   Column coordinate of pixel.                 */
/*----------------------------------------------------------------------------*/
) {
    int i, j, count = 0;

    for(j = n - 1; j < n + 2; j++) {
	if (j < index->hstart || j > index->hend)
	    continue;
	for(i = m - 1; i < m + 2; i++) {
	    if (i < index->vstart[j] || i > index->vend[j])
		continue;
	    if (iMag->data[0][i-index->voffset[j]][j] != 0)
		count++;
	}
    }

    return(count);
}
