#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include <stdio.h>
#include <string.h>
#include "logerr.h"
#if HAVE_PTHREADS
#include <pthread.h>
#endif /* HAVE_PTHREADS */

static logerr_message_poster_t logerr_messenger = NULL;
static logerr_perror_poster_t logerr_perror_messenger = NULL;
static int logerr_abort_threshold = LOGERR_ERR;
#if HAVE_PTHREADS
static pthread_mutex_t logerr_modlock = PTHREAD_MUTEX_INITIALIZER;

#define logerr_lock pthread_mutex_lock (&logerr_modlock)
#define logerr_unlock pthread_mutex_unlock (&logerr_modlock)
#else /* !HAVE_PTHREADS */
#define logerr_lock
#define logerr_unlock
#endif /* !HAVE_PTHREADS */
 
static void
default_messenger (int priority, const char * title, const char * message)
{
  if (title != NULL)
    fprintf(stderr, "%s: ", title);
  else
    fprintf(stderr, "From unknown: ");
  if (message != NULL)
    fprintf(stderr, "%s\n", message);
  else
    fprintf(stderr, "(unspecified message)\n");    
}

static void
default_perror_messenger (int priority, const char * title,
                          const char * message, const char * syserr)
{
  if (title != NULL)
    fprintf(stderr, "%s: ", title);
  else
    fprintf(stderr, "From unknown: ");
  if (message != NULL)
    fprintf(stderr, "%s\n", message);
  else
    fprintf(stderr, "(unspecified message)\n");
  if (syserr != NULL)
    fprintf(stderr, "The system error message was:\n\t%s\n", syserr);
}

void
logerr_vdispatch (int priority, const char * title,
                  const char * fmt, va_list ap)
{
  char message[2048];

  vsnprintf (message, sizeof (message), fmt, ap);
  if (logerr_messenger == NULL)
    default_messenger (priority, title, message);
  else
    (*logerr_messenger) (priority, title, message);
  if (priority <= LOGERR_THRESHOLD_ABORT)
    abort();
}

  
void
logerr_dispatch (int priority, const char * title, const char * fmt, ...)
{
  va_list ap;

  va_start (ap, fmt);
  logerr_vdispatch (priority, title, fmt, ap);
  va_end (ap);
}

void
logerr_vperror (int priority, int errnum, const char * title, const char * fmt, va_list ap)
{
  char message[2048];

  vsnprintf (message, sizeof (message), fmt, ap);
  if (logerr_perror_messenger == NULL)
    default_perror_messenger (priority, title, message, strerror(errnum));
  else
    (*logerr_perror_messenger) (priority, title, message, strerror(errnum));
  if (priority <= LOGERR_THRESHOLD_ABORT)
    abort();
}
  
void
logerr_perror (int priority, int errnum, const char * title, const char * fmt, ...)
{
  va_list ap;

  va_start (ap, fmt);
  logerr_vperror (priority, errnum, title, fmt, ap);
  va_end (ap);
}

void
logerr_set_abort_threshold (int threshold)
{
  logerr_lock;
  logerr_abort_threshold = threshold;
  logerr_unlock;
}

void
logerr_set_messenger (logerr_message_poster_t messenger)
{
  logerr_lock;
  logerr_messenger = messenger;
  logerr_unlock;
}

void
logerr_set_perror_messenger (logerr_message_poster_t poster)
{
  logerr_lock;
  logerr_perror_messenger = poster;
  logerr_unlock;
}
