#ifndef LOGERR_H
#define LOGERR_H 1
#include <stdarg.h>
#include "project_utilities.h"

BEGIN_C_DECLARATIONS

/* Based partly on ideas lifted from the error logging for GLib (the core
   library for GTK+ and GNOME). */

typedef enum
  {
    LOGERR_EMERG   = 0,
    LOGERR_ALERT   = 1,
    LOGERR_CRIT    = 2,
    LOGERR_ERR     = 3,
    LOGERR_WARNING = 4,
    LOGERR_NOTICE  = 5,
    LOGERR_INFO    = 6,
    LOGERR_DEBUG   = 7
  } logerr_priority;

#ifndef LOGERR_THRESHOLD_REPORT
#define LOGERR_THRESHOLD_REPORT LOGERR_DEBUG
#endif

#ifndef LOGERR_THRESHOLD_ABORT
#define LOGERR_THRESHOLD_ABORT LOGERR_ERR
#endif

typedef void (*logerr_message_poster_t) (int priority, const char * title,
                                         const char * message);
typedef void (*logerr_perror_poster_t) (int priority, const char * title,
                                        const char * message, const char * syserr);

void logerr_vdispatch (int priority, const char * title,
                       const char * fmt, va_list ap);
void logerr_dispatch (int priority, const char * title,
                      const char * fmt, ...);

void logerr_vperror (int priority, int errnum, const char * title,
                     const char * fmt, va_list ap);

void logerr_perror (int priority, int errnum, const char * title,
                    const char * fmt, ...);

#if HAVE_C99_VARARGS
#define logerr(priority, title, ...) do { \
    if (priority <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (priority, title, __VA_ARGS__); \
  } while (0)
#define logerrno(priority, errnum, title, ...) do { \
    if (priority <= LOGERR_THRESHOLD_REPORT) \
      logerr_perror (priority, errnum, title, __VA_ARGS__); \
  } while (0)
#define logerr_emerg(title, ...) do { \
    if (LOGERR_EMERG <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_EMERG, title, __VA_ARGS__); \
  } while (0)
#define logerr_alert(title, ...) do { \
    if (LOGERR_ALERT <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_ALERT, title, __VA_ARGS__); \
  } while (0)
#define logerr_crit(title, ...) do { \
    if (LOGERR_CRIT <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_CRIT, title, __VA_ARGS__); \
  } while (0)
#define logerr_err(title, ...) do { \
    if (LOGERR_ERR <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_ERR, title, __VA_ARGS__); \
  } while (0)
#define logerr_warning(title, ...) do { \
    if (LOGERR_WARNING <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_WARNING, title, __VA_ARGS__); \
  } while (0)
#define logerr_notice(title, ...) do { \
    if (LOGERR_NOTICE <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_NOTICE, title, __VA_ARGS__); \
  } while (0)
#define logerr_info(title, ...) do { \
    if (LOGERR_INFO <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_INFO, title, __VA_ARGS__); \
  } while (0)
#define logerr_debug(title, ...) do { \
    if (LOGERR_DEBUG <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_DEBUG, title, __VA_ARGS__); \
  } while (0)
#elif HAVE_OLDGCC_VARARGS
#define logerr(priority, title, fmt...) do { \
    if (priority <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (priority, title, fmt); \
  } while (0)
#define logerrno(priority, errnum, title, fmt...) do { \
    if (priority <= LOGERR_THRESHOLD_REPORT) \
      logerr_perror (priority, errnum, title, fmt); \
  } while (0)
#define logerr_emerg(title, fmt...) do { \
    if (LOGERR_EMERG <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_EMERG, title, fmt); \
  } while (0)
#define logerr_alert(title, fmt...) do { \
    if (LOGERR_ALERT <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_ALERT, title, fmt); \
  } while (0)
#define logerr_crit(title, fmt...) do { \
    if (LOGERR_CRIT <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_CRIT, title, fmt); \
  } while (0)
#define logerr_err(title, fmt...) do { \
    if (LOGERR_ERR <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_ERR, title, fmt); \
  } while (0)
#define logerr_warning(title, fmt...) do { \
    if (LOGERR_WARNING <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_WARNING, title, fmt); \
  } while (0)
#define logerr_notice(title, fmt...) do { \
    if (LOGERR_NOTICE <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_NOTICE, title, fmt); \
  } while (0)
#define logerr_info(title, fmt...) do { \
    if (LOGERR_INFO <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_INFO, title, fmt); \
  } while (0)
#define logerr_debug(title, fmt...) do { \
    if (LOGERR_DEBUG <= LOGERR_THRESHOLD_REPORT) \
      logerr_dispatch (LOGERR_DEBUG, title, fmt); \
  } while (0)
#else /* we can't use varadic preprocessor macros */

static void
logerr (int priority, const char * title, const char * fmt, ...)
{
  va_list ap;

  if (priority <= LOGERR_THRESHOLD_REPORT)
    {
      va_start (ap, fmt);
      logerr_vdispatch (priority, title, fmt, ap);
      va_end (ap);
    }
}

static void
logerrno (int priority, int errnum, const char * title, const char * fmt, ...)
{
  va_list ap;

  if (priority <= LOGERR_THRESHOLD_REPORT)
    {
      va_start (ap, fmt);
      logerr_vperror (priority, errnum, title, fmt, ap);
      va_end (ap);
    }
}

#define LOGERR_MAKE_EASY(NICENAME, UGLYCODE) static void \
logerr_##NICENAME (const char * title, const char * fmt, ...) \
{ \
  va_list ap; \
  if (UGLYCODE <= LOGERR_THRESHOLD_REPORT) \
    { \
      va_start (ap, fmt); \
      logerr_vdispatch (UGLYCODE, title, fmt, ap); \
      va_end (ap); \
    } \
}

LOGERR_MAKE_EASY (emerg, LOGERR_EMERG)
LOGERR_MAKE_EASY (alert, LOGERR_ALERT)
LOGERR_MAKE_EASY (crit, LOGERR_CRIT)
LOGERR_MAKE_EASY (err, LOGERR_ERR)
LOGERR_MAKE_EASY (warning, LOGERR_WARNING)
LOGERR_MAKE_EASY (notice, LOGERR_NOTICE)
LOGERR_MAKE_EASY (info, LOGERR_INFO)
LOGERR_MAKE_EASY (debug, LOGERR_DEBUG)

#endif /* !HAVE_*_VARARGS */

#define logerr_assert(priority, title, condition) do { \
    if (priority <= LOGERR_THRESHOLD_REPORT) { \
        if (!(condition)) \
          logerr_dispatch(priority, title, \
                          "file %s: line %d: assertion (%s) failed", \
                          __FILE__, __LINE__, (#condition)); \
      } \
  } while (0);

void logerr_set_abort_threshold (int threshold);
void logerr_set_messenger (logerr_message_poster_t messenger);
void logerr_set_perror_messenger (logerr_message_poster_t poster);

#if HAVE_C99_FUNC
#define LOGERR_FUNC __func__
#elif HAVE_PRETTY_FUNCTION
#define LOGERR_FUNC __PRETTY_FUNCTION__
#else
#define LOGERR_FUNC NULL
#endif /* HAVE_*_FUNCTION */

END_C_DECLARATIONS

#endif /* !LOGERR_H */

