#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        <stdio.h>
#include        <sadie.h>
#include        "mosaic.h"
#if WITH_DMALLOC
#include <dmalloc.h>
#endif /* WITH_DMALLOC */


/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1999 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-Programmer Information-----------------------------------------------------*/
/*                                                                            */
/*   This function imports an image subsection from a mosaic disk file into   */
/*   main memory.                                                             */
/*   It is based on import.c						      */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function reads an image from disk into main memory.                 */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void MOSAIC_IMPORT (
MOSAIC_INDEX *index, /*  I   Index information for MOSAIC_IMAGE file.         */
char *name,          /*  I   String, containing the name of the disk file.    */
int jbgn,	     /*  I   Index of the first line to be read.	      */
int kbgn,	     /*  I   Index of the first pixel to be read.	      */
int jinc,	     /*  I   Line increment during sampling.		      */
int kinc,	     /*  I   Pixels/line increment during sampling.	      */
int jend,	     /*  I   Index of the last line to be read. 	      */
int kend,	     /*  I   Index of the last pixel to be read.	      */
IMAGE **out     /*  O   Address of a pointer to the output image.        */
/*----------------------------------------------------------------------------*/
) { 
    register int j, k;
    char  msg[SLEN];
    int lines, pixels;
    int outj, outk, nlin, npix;
    int nbytes;
    PIXEL_BYTE *column=NULL, *ptr=NULL;
    FILE *fp=NULL;

    /* if (TIMES) TIMING(T_MOSAIC_IMPORT); */

/*    printf("Importing mosaic of image of size %dx%d from %d,%d to %d,%d with increment %d,%d\n",index->npix,index->nlin,kbgn,jbgn,kend,jend,kinc,jinc); */

    /* check input */
    if (!CHECKIMG_MOSAIC_INDEX(index)) {
        MESSAGE('E'," Can't identify mosaic index.");
        goto the_end;
    } else if (0 > jbgn  ||  jend >= index->nlin) {
        MESSAGE('E'," Image to be sampled must be located completely inside input image in the line dimension.");
        goto the_end;
    } else if (0 > kbgn  ||  kend >= index->npix) {
        MESSAGE('E'," Image to be sampled must be located completely inside input image in the pixel dimension.");
        goto the_end;
    } else if (jinc <= 0) {
        MESSAGE('E'," Line subsample increment must be greater than zero.");
        goto the_end;
    } else if (kinc <= 0) {
        MESSAGE('E'," Pixel subsample increment must be greater than zero.");
        goto the_end;
    } else if (jbgn > jend) {
        MESSAGE('E'," Number of lines to be spanned must be greater than zero.");
        goto the_end;
    } else if (kbgn > kend) {
        MESSAGE('E'," Number of pixels/line to be spanned must be greater than zero.");
        goto the_end;
    }

    /* open mosaic image file */
    if (!(fp=fopen(name,"r"))) {
        sprintf(msg," Can't open file %s.",name);
        MESSAGE('E',msg);
        goto the_end;
    }

    /* Set file pointer to first byte after header */
    fseek(fp, (long) 512, SEEK_SET);


    /* create image of appropriate size */
    lines  = max(1,(jend-jbgn+1)/jinc);
    pixels = max(1,(kend-kbgn+1)/kinc);
    if (!CHECKIMG(*out)) GETMEM(1,lines,pixels,out);
    if (!*out) goto the_end;
    (*out)->nbit = 8;

    /* read the image data from file */
    nbytes = (int)index->height_max;
    if ((column=(PIXEL_BYTE *)malloc((long)index->height_max * sizeof(PIXEL_BYTE))) == NULL) {
	MESSAGE('E'," Memory request failed.");
	goto the_end;
    }	
    if (fseek(fp,(long)(nbytes * (long)kbgn),SEEK_CUR) != 0) {
	sprintf(msg," Can't skip to column %d in file %s.",kbgn,name);
	MESSAGE('E',msg);
	goto the_end;
    }
    for (k = kbgn + (kinc-1), outk=0; outk<pixels; k += kinc, outk++) {
	if (fseek(fp,(long)(nbytes * (long)(kinc - 1)),SEEK_CUR) != 0) {
	    sprintf(msg," Can't column subsample by %d in file %s.",kinc,name);
	    MESSAGE('E',msg);
	    goto the_end;
	}
	if (fread(column,nbytes,1,fp) != 1) {
	    sprintf(msg," Can't read image data from file %s.",name);
	    MESSAGE('E',msg);
	    goto the_end;
	}
	for (j = jbgn, ptr=column, outj=0; outj < lines; j+=jinc, outj++) {
	    if ((j < index->voffset[k]) || (j > index->voffset[k] + index->vheight[k])) {
		(*out)->data[0][outj][outk] = (PIXEL) 0.0;
	    } else {
		(*out)->data[0][outj][outk] = (PIXEL) *(ptr + (j - index->voffset[k]));
	    }
	}
    }

    
    if (NAMES) {
        RANGE(*out);
        MESSAGE('I',"");
        MESSAGE('I',"MOSAIC_IMPORT");
        MESSAGE('I',"");
        sprintf(msg," Input file:                                 %s",name);
        MESSAGE('I',msg); 
	MESSAGE('I'," Input format:                               8-bit BIP in MOSAIC_IMAGE format");
        MESSAGE('I',msg);
        sprintf(msg," Number of lines, pixels:                    %d, %d",nlin,npix);
        MESSAGE('I',msg);
        sprintf(msg," Number of bits/pixel/band:                  %d",8);
        MESSAGE('I',msg);
        sprintf(msg," Subsampled from line, pixel:                %d, %d",jbgn+1,kbgn+1);
        MESSAGE('I',msg);
        sprintf(msg," Subsampled to line, pixel:                  %d, %d",jend+1,kend+1);
        MESSAGE('I',msg);
        sprintf(msg," Subsample increment: lines, pixels:         %d, %d",jinc,kinc);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* close image file */
    the_end:
    if (fp) {
	if (fclose(fp)) {
	    sprintf(msg," Can't close file %s.",name);
	    MESSAGE('W',msg);
	}
    }
    if (column)  free(column);
    if (TIMES) TIMING(T_EXIT);
}
