/*
 * tkCamera.c -- This file creates a "camera" image that is used to display
 *			a live image from an SDV digital video camera in a
 *                      Tcl/Tk interface.  This file is loosely modeled after
 *                      tkImgPhoto.c which is contained in the generic directory
 *                      of the Tk source code tree.  A single photo image is created,
 *                      but the data in the image is updated each time a new frame is
 *                      captured.  Thus the "camera" image can be imbedded in a canvas
 *                      and later updated without touching the canvas.
 */



#include <tk.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include "cameradevice.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <iostream>
#define MAXSTR  200
#define MAXNAME 100

#ifdef __cplusplus
namespace std
{
#endif 

/* Define a global camera device that is accessable from all procedures.
 * This assumes that only one camera will be used in the application.
 */
Cameradevice *globalcamdev=NULL;



/*
 * The Camera typedef is a data structure that defines one instance
 * of the widget.  There might be many instances in one interface.
 */
typedef struct {
	Tk_Window tkwin;			/* The window for the widget (not used, only for storing configuration) */
        char name[MAXNAME];             /* The name of the camera widget */
        char photo[MAXNAME];            /* The name of photo image to hold the video sequence */
        Tk_PhotoHandle imageHandle; /* The tk imageHandle in which camera images are displayed */
	Tcl_Interp *interp;			/* The interpreter of the widget */
	Tcl_Command widgetCmd;		/* Token for message's widget command. */
	Tk_TimerToken token;		/* Handle on periodic callbacks */
	Cameradevice *camdev;		/* Camera device for capturing frames */
	int paused;					/* Indicates if the camera device is paused */
	int storetofile;			/* Indicates that the current image should be stored to filename */
	char* filename;				/* Indicates name of file into which an image will be stored */
	int unit;					/* Indicates which sdv unit */
	int numbufs;				/* Indicates number of buffers for speed */
	int loopcount;				/* Indicates number of images to capture */
	int picture_timeout;		/* Indicates acquisition timeout */
	int hoff;					/* Indicates offset from left of camera */
	int voff;					/* Indicates offset from top of camera */
	int deci;					/* Indicates decimation ratio */
	int width;					/* in pixels */
	int height;					/* in pixels */
	int configwidth;
	int configheight;
	int flags;					/* Flags defined below */
} Camera;





/*
 * Flag bits.
 *  REDRAW_PENDING  - a call to CameraGrab has been queued already.
 *  GOT_FOCUS   - the input focus is directed at the camera.
 *  TICKING - periodic callback is enabled.
 */
#define REDRAW_PENDING	0x1
#define GOT_FOCUS	0x2
#define TICKING		0x4





/*
 * These configuration specs are used to automatically process
 * arguments to the command that creates a Camera and to the
 * configure operation that changes a widget.  They define the
 * command line flag, its type, its resource name, its class name,
 * and its default value.  Tk_Offset is a macro that gets the
 * address of the structure member so the Camera struct can
 * be modified directly.
 */

static Tk_ConfigSpec configSpecs[] =
{
	{TK_CONFIG_PIXELS, "-width", "width", "Width",
	 "2", Tk_Offset(Camera, configwidth), 0},
	{TK_CONFIG_PIXELS, "-height", "height", "Height",
	 "2", Tk_Offset(Camera, configheight), 0},
	{TK_CONFIG_PIXELS, "-pause", "pause", "Pause",
	 "0", Tk_Offset(Camera, paused), 0},
	{TK_CONFIG_STRING, "-filename", "filename", "Filename",
	 "none", Tk_Offset(Camera, filename), 0},
	{TK_CONFIG_PIXELS, "-storetofile", "storetofile", "Storetofile",
	 "0", Tk_Offset(Camera, storetofile), 0},
	{TK_CONFIG_PIXELS, "-unit", "unit", "Unit",
	 "0", Tk_Offset(Camera, unit), 0},
	{TK_CONFIG_PIXELS, "-numbufs", "numbufs", "Numbufs",
	 "10", Tk_Offset(Camera, numbufs), 0},
	{TK_CONFIG_PIXELS, "-loopcount", "loopcount", "Loopcount",
	 "-1", Tk_Offset(Camera, loopcount), 0},
	{TK_CONFIG_PIXELS, "-timeout", "timeout", "Timeout",
	 "-1", Tk_Offset(Camera, picture_timeout), 0},
	{TK_CONFIG_PIXELS, "-hoffset", "hoffset", "Hoffset",
	 "0", Tk_Offset(Camera, hoff), 0},
	{TK_CONFIG_PIXELS, "-voffset", "voffset", "Voffset",
	 "0", Tk_Offset(Camera, voff), 0},
	{TK_CONFIG_PIXELS, "-deci", "deci", "Deci",
	 "2", Tk_Offset(Camera, deci), 0},
	{TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
	 (char *) NULL, 0, 0}
};





/*
 * Forward declarations.  _ANSI_ARGS_ is a macro that checks to
 * see if we are being compiled with an ANSI C compiler before
 * it puts the parameter types into the declaration.
 */
/*
 * CameraFree --
 *  Clean up after a camera is destroyed.
 *  This frees all the resources associated with the camera.
 */
static void
CameraFree(char *blockPtr)
{
	Camera* cameraPtr = (Camera*) blockPtr;
	Tcl_Obj* cmd = Tcl_NewStringObj("image delete ", -1);

	cerr << "Starting deletion of " << cameraPtr->name << endl;
	if (cameraPtr == NULL) return;
	if (cameraPtr->photo != NULL) {
          cerr << "Deletion of " << cameraPtr->photo << " attempted" << endl;
	  Tcl_AppendStringsToObj(cmd, cameraPtr->photo, 0);
	  if (Tcl_EvalObj(cameraPtr->interp, cmd) != TCL_OK) {
	    cerr << "Deletion of " << cameraPtr->photo << " result: " << Tcl_GetStringResult(cameraPtr->interp) << endl;
	  }
	}
	if (cameraPtr->camdev) {
	  delete cameraPtr->camdev;
	  cameraPtr->camdev = NULL;
	}
	Tk_DestroyWindow(cameraPtr->tkwin);
	if (cameraPtr) {
	  Tcl_Free((char*) cameraPtr);
	}
}


/*
 * CameraGrab --
 *  This is called to grab an image with the camera.
 */
static void
CameraGrab(ClientData clientData)
{
	Camera *cameraPtr = (Camera *) clientData;
    Tk_PhotoImageBlock block;

    Tk_Preserve(clientData);

    /*
     * Create an image block for a grayscale image
     */
    block.pixelSize = 1;
    block.offset[0] = 0;
    block.offset[1] = 0;
    block.offset[2] = 0;
    block.width = cameraPtr->width;
    block.height = cameraPtr->height;
    block.pitch = block.pixelSize * block.width;
    block.pixelPtr = 0x0;



	cameraPtr->flags &= ~(REDRAW_PENDING | TICKING);


	/*
	 * Update the image
	 */
	if (!cameraPtr->paused) {
		block.pixelPtr = cameraPtr->camdev->acquire();
		if (block.pixelPtr) {
#if (TK_MAJOR_VERSION > 8) || ((TK_MAJOR_VERSION == 8) && (TK_MINOR_VERSION > 3))
                        Tk_PhotoPutBlock(cameraPtr->imageHandle, &block, 0, 0, block.width, block.height, TK_PHOTO_COMPOSITE_SET);
#else /* (TK_MAJOR_VERSION < 8) || ((TK_MAJOR_VERSION == 8) && (TK_MINOR_VERSION <= 3)) */
                        Tk_PhotoPutBlock(cameraPtr->imageHandle, &block, 0, 0, block.width, block.height);
#endif
		}
	}



	/*
	 * Queue another call to ourselves.  The rate at which this
	 * is done could be optimized.
	 */
	while (Tk_DoOneEvent(TK_DONT_WAIT) != 0);
    
	if (cameraPtr->loopcount != 0) {
		cameraPtr->loopcount--;
		cameraPtr->token = Tk_CreateTimerHandler(10, CameraGrab,
												 (ClientData) cameraPtr);
		cameraPtr->flags |= TICKING;
	} else {
	      Tk_CancelIdleCall(CameraGrab, (ClientData) cameraPtr);
	      cameraPtr->flags &= ~REDRAW_PENDING;
	      if (cameraPtr->token) {
		    Tk_DeleteTimerHandler(cameraPtr->token);
		    cameraPtr->token = 0;
		    cameraPtr->flags &= ~TICKING;
	      }
	      if (cameraPtr->widgetCmd) {
		    Tcl_DeleteCommandFromToken(cameraPtr->interp, cameraPtr->widgetCmd);
		    cameraPtr->widgetCmd = 0;
		    Tk_EventuallyFree((ClientData) cameraPtr, CameraFree);
	      }
	}

	Tk_Release(clientData);
}


/*
 * CameraConfigure --
 *  This processes command line arguments defined in the
 *  TkConfigSpecs array.  This is called when a widget is first
 *  created or when its configuration is changed via the
 *  configure operation.
 *  It returns a standard Tcl result.
 */
static int
CameraConfigure(Tcl_Interp * interp, Camera * cameraPtr, int argc, const char **argv, int flags)
{
	/*
	 * Tk_ConfigureWidget does all the work of parsing the
	 * arguments and looking for things in the resource database.
	 */
	if (Tk_ConfigureWidget(interp, cameraPtr->tkwin, configSpecs,
						 argc, argv, (char *) cameraPtr, flags) != TCL_OK) {
		return TCL_ERROR;
	}
	/*
	 * Initialize the camera device
	 */
	if (!cameraPtr->camdev) {
		printf("initializing the camera...\n");
		cameraPtr->camdev = new Cameradevice(cameraPtr->unit, cameraPtr->picture_timeout,
		cameraPtr->numbufs, cameraPtr->hoff, cameraPtr->voff, 1, cameraPtr->deci);
		cameraPtr->width = cameraPtr->camdev->width();
		cameraPtr->height = cameraPtr->camdev->height();
	}
	
	globalcamdev = cameraPtr->camdev;

	
	
	//if (cameraPtr->paused) {
	//	cameraPtr->camdev->enable_live(0);
	//} else {
	//	cameraPtr->camdev->enable_live(1);
	//}


	/*
	 * Set up a call to grab a frame
	 */
	if ((cameraPtr->tkwin != NULL) && !(cameraPtr->flags & REDRAW_PENDING)) {
		Tk_DoWhenIdle(CameraGrab, (ClientData) cameraPtr);
		cameraPtr->flags |= REDRAW_PENDING;
	}



	return TCL_OK;
}

/*
 * CameraInstanceCmd --
 *  This procedure implements the operations on an instance
 *  of the widget.   It is a command procedure for the Tcl
 *  command that is the same as the name of the widget.
 *  This returns a standard Tcl result.
 */
static int
CameraInstanceCmd(ClientData clientData, Tcl_Interp * interp, int argc, const char **argv)
{
	Camera *cameraPtr = (Camera *) clientData;
	int result = TCL_OK;
	char c;
	int length;

	if (argc < 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
						 argv[0], " option ?arg arg ...?\"", (char *) NULL);
		return TCL_ERROR;
	}
	Tk_Preserve(clientData);
	c = argv[1][0];
	length = strlen(argv[1]);
	if ((c == 'c') && (strncmp(argv[1], "cget", length) == 0)
		&& (length >= 2)) {
		if (argc != 3) {
			Tcl_AppendResult(interp, "wrong # args: should be \"",
							 argv[0], " cget option\"",
							 (char *) NULL);
			Tk_Release(clientData);
			return TCL_ERROR;
		}
		result = Tk_ConfigureValue(interp, cameraPtr->tkwin, configSpecs,
								   (char *) cameraPtr, argv[2], 0);
	} else if ((c == 'c') && (strncmp(argv[1], "configure", length) == 0)
			   && (length >= 2)) {
		if (argc == 2) {
			/*
			 * Return info about all attributes.
			 */
			result = Tk_ConfigureInfo(interp, cameraPtr->tkwin, configSpecs,
									  (char *) cameraPtr, (char *) NULL, 0);
		} else if (argc == 3) {
			/*
			 * Return info about one attribute, like cget.
			 */
			result = Tk_ConfigureInfo(interp, cameraPtr->tkwin, configSpecs,
									  (char *) cameraPtr, argv[2], 0);
		} else {
			/*
			 * Change one or more attributes.
			 */
			result = CameraConfigure(interp, cameraPtr, argc - 2, argv + 2,
									 TK_CONFIG_ARGV_ONLY);
		}
	} else {
		Tcl_AppendResult(interp, "bad option \"", argv[1],
						 "\":  must be cget, configure, position, or size",
						 (char *) NULL);
		Tk_Release(clientData);
		return TCL_ERROR;
	}
	Tk_Release(clientData);
	return result;
}

/*
 * CameraCmd --
 *  This is the command procedure for the "camera" Tcl command.
 *  It creates an instance of the widget.  As a side effect, a new
 *  Tcl command is created that operates on the new instance.
 *  This returns TCL_OK or TCL_ERROR if the widget could not be
 *  created.
 */
int
CameraCmd(ClientData clientData, Tcl_Interp * interp, int argc, const char **argv)
{
	Tk_Window main = (Tk_Window) clientData;
	Camera *cameraPtr;
        Tk_PhotoHandle imageHandle;
        char tclstr[MAXSTR];
        char* defaultphotoname = "TK_VIDEO_PHOTO_IMAGE";
        char* photoname = defaultphotoname;
	Tk_Window tkwin;

	
	if (argc < 2) {
		Tcl_AppendResult(interp, "wrong # args:  should be \"",
						 argv[0], " pathName ?options?\"", (char *) NULL);
		return TCL_ERROR;
	}
    
	tkwin = Tk_CreateWindow(interp, main, argv[1], (char *) NULL);
	if (tkwin == NULL) {
                Tcl_SetResult(interp, "ERROR: Video window not created...\n", TCL_STATIC);
		return TCL_ERROR;
	}
	Tk_SetClass(tkwin, "Camera");

	imageHandle = Tk_FindPhoto(interp, photoname);
	if (imageHandle != NULL) {
                Tcl_SetResult(interp, "Error: Video Photo Image already exists in this application.", TCL_STATIC);
		return TCL_ERROR;
	}
	
	snprintf(tclstr, MAXSTR, "image create photo %s", photoname);
	if( Tcl_Eval(interp, tclstr) != TCL_OK ) {
	  return TCL_ERROR;
	}

	imageHandle = Tk_FindPhoto(interp, photoname);
	if (imageHandle == NULL) {
                Tcl_SetResult(interp, "Error: Video Photo Image could not be created.", TCL_STATIC);
		return TCL_ERROR;
	}

	/*
	 * Allocate and initialize the widget record.
	 */

	cameraPtr = (Camera *) Tcl_Alloc(sizeof (Camera));
	cameraPtr->tkwin = tkwin;
	snprintf(cameraPtr->name, MAXNAME, argv[1]);
	snprintf(cameraPtr->photo, MAXNAME, photoname);
	cameraPtr->imageHandle = imageHandle;
	cameraPtr->interp = interp;
	cameraPtr->token = NULL;
	cameraPtr->camdev = NULL;
	cameraPtr->paused = 0;
	cameraPtr->storetofile = 0;
	cameraPtr->filename = NULL;
	cameraPtr->unit = 0;
	cameraPtr->numbufs = 10;
	cameraPtr->loopcount = -1;
	cameraPtr->picture_timeout = -1;
	cameraPtr->hoff = 0;
	cameraPtr->voff = 0;
	cameraPtr->deci = 2;
	cameraPtr->width = 0;
	cameraPtr->height = 0;
	cameraPtr->configwidth = 0;
	cameraPtr->configheight = 0;
	cameraPtr->flags = 0;


	/*
	 * Create a Tcl command that operates on the widget.
	 */
	cameraPtr->widgetCmd = Tcl_CreateCommand(interp, cameraPtr->name, CameraInstanceCmd,
						 (const ClientData) cameraPtr, 0);



	/*
	 * Parse the command line arguments.
	 */
        int shifted_argc = argc - 2;
        const char** shifted_argv = argv + 2;
	if (CameraConfigure(interp, cameraPtr, shifted_argc, shifted_argv, 0) != TCL_OK) {
		printf("ERROR: Could not configure widget...\n");
		Tk_DestroyWindow(cameraPtr->tkwin);
		return TCL_ERROR;
	}
    
    
        Tcl_SetResult(interp, cameraPtr->photo, TCL_VOLATILE);
	return TCL_OK;
}


#ifdef __cplusplus
extern "C" {
#endif 

int tkCamera_Init(Tcl_Interp *interp) {

  Tcl_CreateCommand(interp, "camera", CameraCmd, 
                    (ClientData)Tk_MainWindow(interp), 0);
  return TCL_OK;
}

#ifdef __cplusplus
} // extern "C"
} // namespace std
#endif 
