#if HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include "trees.h"
#if WITH_DMALLOC
#include <dmalloc.h>
#endif /* WITH_DMALLOC */

static PIXEL firstdiff_x[3][3] = { { -1.0,  0.0,  1.0 }, 
                                   { -2.0,  0.0,  2.0 }, 
                                   { -1.0,  0.0,  1.0 } };

static PIXEL firstdiff_y[3][3] = { {  1.0,  2.0,  1.0 }, 
                                   {  0.0,  0.0,  0.0 }, 
                                   { -1.0, -2.0, -1.0 } };


/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/* This function detects tree rings.                                          */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void TREES_TREERINGMAG (
IMAGE  *in,      /*  I   Pointer to the input image.                          */
LOCAL_INDEX *localindex,  		      
double sigma,    /*  I   Std. deviation of gaussian smoothing filter.         */
int    size,     /*  I   Size of gaussian smoothing filter.                   */
IMAGE **mag,     /*  O   Address of a pointer to the output image             */
                 /*      containing the gradient magnitude.                   */
IMAGE **dir      /*  O   Address of a pointer to the output image             */
                 /*      containing the gradient direction.                   */
/*----------------------------------------------------------------------------*/
) { register int j, k, n;
    char msg[SLEN];
    IMAGE *gaussian, *smoothed;
    IMAGE *FoG_mag, *FoG_dir;
    PIXEL *psf=0, factor;

    /* if (TIMES) TIMING(T_TREERINGMAG); */
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"TREERINGMAG");
        MESSAGE('I',"");
        sprintf(msg," Input image:                       %s",in->text);
        MESSAGE('I',msg);
        sprintf(msg," Gaussian smoothing filter size:    %d",size);
        MESSAGE('I',msg);
        sprintf(msg,"             Standard Deviation:    %f",sigma);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* check input */
    if (!CHECKIMG(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (in->nbnd > 1) {
        MESSAGE('E'," Image must be single-band.");
        goto the_end;
    }

    if (size > 0) {
        /* create gaussian smoothing filter */
        CREATEGAUSS(size, sigma, &gaussian);
        if (CHECKIMG(gaussian)) {
            sprintf(gaussian->text, "%dx%d Gaussian -- sigma=%f", size,size,sigma);
        } else {
            MESSAGE('E'," Can't identify gaussian filter.");
            goto the_end;
        }

        /* perform gaussian smoothing on input image */
        if (!(psf = (PIXEL *) malloc(size*size*sizeof(PIXEL)))) {
            MESSAGE('E'," Memory allocation failed.");
            goto the_end;
        }
        for (n=0, j=0, factor=0.0; j<size; j++) {
            for (k=0; k<size; k++) {
                psf[n] = gaussian->data[0][j][k];
                factor += psf[n++];
            }
        }
        for (n=0, j=0; j<size; j++) {
            for (k=0; k<size; k++) {
                psf[n++] /= factor;
            }
        }
        TREES_SCONVL(in, localindex, psf, size, size, &smoothed);
    } else {
        IMGCOPY(in,&smoothed);
    }

    if (CHECKIMG(smoothed)) {
        sprintf(smoothed->text, "Smoothed image");
    } else {
        MESSAGE('E'," Can't identify smoothed image.");
        goto the_end;
    }
    
    /* create the First Order Gaussian filtered image */
    TREES_GRADIENT(smoothed, localindex, (PIXEL *)firstdiff_y, (PIXEL *)firstdiff_x, 3, &FoG_mag, &FoG_dir);
    
    *mag = FoG_mag;
    *dir = FoG_dir;
    
    the_end:

    if (psf) free(psf);
    if (CHECKIMG(gaussian))  RELIMG(&gaussian);
    if (CHECKIMG(smoothed))  RELIMG(&smoothed);
    /* if (TIMES) TIMING(T_EXIT); */

}



